import numpy as np
import h5py

from fluidsim.base.output.spect_energy_budget import (
    SpectralEnergyBudgetBase,
    cumsum_inv,
    inner_prod,
)

from .normal_mode import NormalModeDecomposition, NormalModeDecompositionModified


class SpectralEnergyBudgetSW1LBase(SpectralEnergyBudgetBase):
    """Save and plot spectral energy budgets."""

    def __init__(self, output):

        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f

        super().__init__(output)

    def _checksum_stdout(self, debug=False, **kwargs):
        if debug is True:
            for key, value in list(kwargs.items()):
                if mpi.nb_proc > 1:
                    value_g = mpi.comm.gather(value, root=0)
                else:
                    value_g = value

                if mpi.rank == 0:
                    print(
                        ("sum({0}) = {1:9.4e} ; sum(|{0}|) = {2:9.4e}").format(
                            key, np.sum(value_g), np.sum(np.abs(value_g))
                        )
                    )


class SpectralEnergyBudgetMSW1L(SpectralEnergyBudgetSW1LBase):
    def compute(self):
        """compute spectral energy budget the one time."""
        oper = self.sim.oper

        try:
            state_spect = self.sim.state.state_spect
            ux_fft = state_spect.get_var("ux_fft")
            uy_fft = state_spect.get_var("uy_fft")
            eta_fft = state_spect.get_var("eta_fft")
        except ValueError:
            state = self.sim.state
            ux_fft = state.get_var("ux_fft")
            uy_fft = state.get_var("uy_fft")
            eta_fft = state.get_var("eta_fft")

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        urx_fft, ury_fft = oper.vecfft_from_rotfft(rot_fft)
        del rot_fft
        urx = oper.ifft2(urx_fft)
        ury = oper.ifft2(ury_fft)

        q_fft, div_fft, a_fft = self.oper.qdafft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        udx_fft, udy_fft = oper.vecfft_from_divfft(div_fft)

        if self.params.f != 0:
            ugx_fft, ugy_fft, etag_fft = self.oper.uxuyetafft_from_qfft(q_fft)
            uax_fft, uay_fft, etaa_fft = self.oper.uxuyetafft_from_afft(a_fft)
        # velocity influenced by linear terms
        # u_infl_lin_x = udx_fft + uax_fft
        # u_infl_lin_y = udy_fft + uay_fft

        # compute flux of Charney PE
        Fq_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, q_fft)

        transferCPE_fft = inner_prod(q_fft, Fq_fft)
        del (q_fft, Fq_fft)

        transfer2D_CPE = self.spectrum2D_from_fft(transferCPE_fft)
        del transferCPE_fft

        #         print(
        # ('sum(transfer2D_CPE) = {0:9.4e} ; sum(abs(transfer2D_CPE)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_CPE),
        # np.sum(abs(transfer2D_CPE)))
        # )

        Feta_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, eta_fft)
        transferEA_fft = self.c2 * inner_prod(eta_fft, Feta_fft)
        del Feta_fft
        transfer2D_EA = self.spectrum2D_from_fft(transferEA_fft)
        del transferEA_fft

        #         print(
        # ('sum(transfer2D_EA) = {0:9.4e} ; sum(abs(transfer2D_EA)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_EA),
        # np.sum(abs(transfer2D_EA)))
        # )

        convA_fft = self.c2 * inner_prod(eta_fft, div_fft)
        convA2D = self.spectrum2D_from_fft(convA_fft)
        del convA_fft

        Fxrr_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, urx_fft)
        Fyrr_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, ury_fft)

        Fxrd_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, udx_fft)
        Fyrd_fft = self.fnonlinfft_from_uxuy_funcfft(urx, ury, udy_fft)

        transferErrr_fft = inner_prod(urx_fft, Fxrr_fft) + inner_prod(
            ury_fft, Fyrr_fft
        )
        transfer2D_Errr = self.spectrum2D_from_fft(transferErrr_fft)
        del transferErrr_fft
        #         print(
        # ('sum(transfer2D_Errr) = {0:9.4e} ; sum(abs(transfer2D_Errr)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Errr),
        # np.sum(abs(transfer2D_Errr)))
        # )

        transferEdrd_fft = inner_prod(udx_fft, Fxrd_fft) + inner_prod(
            udy_fft, Fyrd_fft
        )
        transfer2D_Edrd = self.spectrum2D_from_fft(transferEdrd_fft)
        del transferEdrd_fft
        #         print(
        # ('sum(transfer2D_Edrd) = {0:9.4e} ; sum(abs(transfer2D_Edrd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Edrd),
        # np.sum(abs(transfer2D_Edrd)))
        # )

        Clfromqq = inner_prod(udx_fft, Fxrr_fft) + inner_prod(udy_fft, Fyrr_fft)
        transferEdrr_rrd_fft = (
            Clfromqq
            + inner_prod(urx_fft, Fxrd_fft)
            + inner_prod(ury_fft, Fyrd_fft)
        )
        Clfromqq = self.spectrum2D_from_fft(Clfromqq)
        transfer2D_Edrr_rrd = self.spectrum2D_from_fft(transferEdrr_rrd_fft)
        del transferEdrr_rrd_fft
        #         print(
        # ('sum(transfer2D_Edrr_rrd) = {0:9.4e} ; '
        # 'sum(abs(transfer2D_Edrr_rrd)) = {1:9.4e}'
        # ).format(
        # np.sum(transfer2D_Edrr_rrd),
        # np.sum(abs(transfer2D_Edrr_rrd)))
        # )

        transfer2D_EK = transfer2D_Errr + transfer2D_Edrd + transfer2D_Edrr_rrd
        dict_results = {
            "transfer2D_EK": transfer2D_EK,
            "transfer2D_Errr": transfer2D_Errr,
            "transfer2D_Edrd": transfer2D_Edrd,
            "Clfromqq": Clfromqq,
            "transfer2D_Edrr_rrd": transfer2D_Edrr_rrd,
            "transfer2D_EA": transfer2D_EA,
            "convA2D": convA2D,
            "transfer2D_CPE": transfer2D_CPE,
        }
        # self._checksum_stdout(
        #     EK_GGG=transfer2D_Errr,
        #     EK_GGA=transfer2D_Edrr_rrd,
        #     EK_AAG=transfer2D_Edrd,
        #     EA=transfer2D_EA,
        #     Etot=transfer2D_EK + transfer2D_EA,
        #     debug=False,
        # )
        return dict_results

    def _online_plot_saving(self, dict_results):

        transfer2D_CPE = dict_results["transfer2D_CPE"]
        transfer2D_EK = dict_results["transfer2D_EK"]
        transfer2D_EA = dict_results["transfer2D_EA"]
        convA2D = dict_results["convA2D"]
        khE = self.oper.khE
        PiCPE = cumsum_inv(transfer2D_CPE) * self.oper.deltak
        PiEK = cumsum_inv(transfer2D_EK) * self.oper.deltak
        PiEA = cumsum_inv(transfer2D_EA) * self.oper.deltak
        CCA = cumsum_inv(convA2D) * self.oper.deltak
        self.axe_a.plot(khE + khE[1], PiEK, "r")
        self.axe_a.plot(khE + khE[1], PiEA, "b")
        self.axe_a.plot(khE + khE[1], CCA, "y")
        self.axe_b.plot(khE + khE[1], PiCPE, "g")

    def plot(self, tmin=0, tmax=1000, delta_t=2):

        with h5py.File(self.path_file, "r") as h5file:

            dset_times = h5file["times"]
            dset_khE = h5file["khE"]
            khE = dset_khE[...]
            # khE = khE+khE[1]

            dset_transfer2D_EK = h5file["transfer2D_EK"]
            dset_transfer2D_Errr = h5file["transfer2D_Errr"]
            dset_transfer2D_Edrd = h5file["transfer2D_Edrd"]
            dset_transfer2D_Edrr_rrd = h5file["transfer2D_Edrr_rrd"]
            dset_transfer2D_EA = h5file["transfer2D_EA"]
            dset_convA2D = h5file["convA2D"]
            dset_transfer2D_CPE = h5file["transfer2D_CPE"]

            times = dset_times[...]
            # nt = len(times)

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = """plot fluxes 2D
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            x_left_axe = 0.12
            z_bottom_axe = 0.36
            width_axe = 0.85
            height_axe = 0.57

            size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
            fig, ax1 = self.output.figure_axe(size_axe=size_axe)
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("transfers")
            ax1.set_title("energy flux\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("linear")

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot, delta_i_plot):
                    transferEK = dset_transfer2D_EK[it]
                    transferEA = dset_transfer2D_EA[it]
                    PiEK = cumsum_inv(transferEK) * self.oper.deltak
                    PiEA = cumsum_inv(transferEA) * self.oper.deltak
                    PiE = PiEK + PiEA
                    ax1.plot(khE, PiE, "k", linewidth=1)

            transferEK = dset_transfer2D_EK[imin_plot:imax_plot].mean(0)
            transferEA = dset_transfer2D_EA[imin_plot:imax_plot].mean(0)
            PiEK = cumsum_inv(transferEK) * self.oper.deltak
            PiEA = cumsum_inv(transferEA) * self.oper.deltak
            PiE = PiEK + PiEA

            ax1.plot(khE, PiE, "k", linewidth=2)
            ax1.plot(khE, PiEK, "r", linewidth=2)
            ax1.plot(khE, PiEA, "b", linewidth=2)

            transferEdrr_rrd = dset_transfer2D_Edrr_rrd[imin_plot:imax_plot].mean(
                0
            )
            transferErrr = dset_transfer2D_Errr[imin_plot:imax_plot].mean(0)
            transferEdrd = dset_transfer2D_Edrd[imin_plot:imax_plot].mean(0)

            Pi_drr_rrd = cumsum_inv(transferEdrr_rrd) * self.oper.deltak
            Pi_rrr = cumsum_inv(transferErrr) * self.oper.deltak
            Pi_drd = cumsum_inv(transferEdrd) * self.oper.deltak

            ax1.plot(khE, Pi_drr_rrd, "m:", linewidth=1)
            ax1.plot(khE, Pi_rrr, "m--", linewidth=1)
            ax1.plot(khE, Pi_drd, "m-.", linewidth=1)

            convA2D = dset_convA2D[imin_plot:imax_plot].mean(0)
            CCA = cumsum_inv(convA2D) * self.oper.deltak

            ax1.plot(khE + khE[1], CCA, "y", linewidth=2)

            z_bottom_axe = 0.07
            height_axe = 0.17
            size_axe[1] = z_bottom_axe
            size_axe[3] = height_axe
            ax2 = fig.add_axes(size_axe)
            ax2.set_xlabel("$k_h$")
            ax2.set_ylabel("transfers")
            ax2.set_title("Charney PE flux")
            ax2.set_xscale("log")
            ax2.set_yscale("linear")

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    transferCPE = dset_transfer2D_CPE[it]
                    PiCPE = cumsum_inv(transferCPE) * self.oper.deltak
                    ax2.plot(khE, PiCPE, "g", linewidth=1)

            transferCPE = dset_transfer2D_CPE[imin_plot:imax_plot].mean(0)
            PiCPE = cumsum_inv(transferCPE) * self.oper.deltak

        ax2.plot(khE, PiCPE, "m", linewidth=2)


class SpectralEnergyBudgetSW1L(SpectralEnergyBudgetSW1LBase):
    def __init__(self, output, norm_mode=None):
        if norm_mode is None:
            self.norm_mode = NormalModeDecomposition(output)
        else:
            self.norm_mode = norm_mode

        super().__init__(output)

    def _transfer_keys_coeff(self):
        c2 = self.params.c2
        # TODO: Check again for possible bugs as found in SpectralEnergyBudgetSW1LModified
        keys = {
            "uuu": ["ux_fft", "ux", "px_ux"],  # Quad. K.E. transfer terms
            "uvu": ["ux_fft", "uy", "py_ux"],
            "vuv": ["uy_fft", "ux", "px_uy"],
            "vvv": ["uy_fft", "uy", "py_uy"],
            "eeu": ["px_eta_fft", "eta", "ux"],  # Quad. A.P.E. transfer terms
            "eev": ["py_eta_fft", "eta", "uy"],
            "uud": [
                "ux_fft",
                "ux",
                "div",
            ],  # NonQuad. K.E. - Quad K.E. transfer terms
            "vvd": ["uy_fft", "uy", "div"],
            "dee": ["div_fft", "eta", "eta"],
        }  # NonQuad. K.E. - Quad A.P.E. transfer terms

        coeff = {
            "uuu": -1.0,
            "uvu": -1.0,
            "vuv": -1.0,
            "vvv": -1.0,
            "eeu": 0.5 * c2,
            "eev": 0.5 * c2,
            "uud": -0.5,
            "vvd": -0.5,
            "dee": 0.25 * c2,
        }

        return keys, coeff

    def compute(self):
        """Compute spectral energy budget once for current time."""

        oper = self.oper
        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        # eta_fft = get_var("eta_fft")
        c2 = self.params.c2

        ux = self.sim.state.state_phys.get_var("ux")
        uy = self.sim.state.state_phys.get_var("uy")
        eta = self.sim.state.state_phys.get_var("eta")

        Mx = eta * ux
        My = eta * uy
        Mx_fft = oper.fft2(Mx)
        My_fft = oper.fft2(My)
        oper.dealiasing(Mx_fft, My_fft)
        del (Mx, My)

        self.norm_mode.compute()

        Tq_keys, Tq_coeff = self._transfer_keys_coeff()

        Tq_terms = dict.fromkeys(Tq_keys)
        for key in list(Tq_keys.keys()):
            triad_key_modes, Tq_terms[key] = self.norm_mode.triad_from_keyfftphys(
                *Tq_keys[key]
            )

        Tq_fft = dict.fromkeys(triad_key_modes[0], 0.0)
        n_modes = triad_key_modes[0].shape[0]
        for i in range(n_modes):  # GGG, GGA etc.
            k = triad_key_modes[0][i]
            for j in list(Tq_keys.keys()):  # uuu, uuv etc.
                Tq_fft[k] += np.real(Tq_coeff[j] * Tq_terms[j][i])

        del (Tq_keys, Tq_terms, Tq_coeff)

        # -------------------------
        # Quadratic exchange terms
        # -------------------------
        Cq_keys = {"ue": ["ux_fft", "px_eta_fft"], "ve": ["uy_fft", "py_eta_fft"]}
        Cq_terms = dict.fromkeys(Cq_keys)
        for key in list(Cq_keys.keys()):
            dyad_key_modes, Cq_terms[key] = self.norm_mode.dyad_from_keyfft(
                True, *Cq_keys[key]
            )

        Cq_coeff = {"ue": -c2, "ve": -c2}
        Cq_fft = dict.fromkeys(dyad_key_modes[0], 0.0)
        n_modes = dyad_key_modes[0].shape[0]
        for i in range(n_modes):  # GG, AG, aG, AA
            k = dyad_key_modes[0][i]
            for j in list(Cq_keys.keys()):  # ue, ve
                Cq_fft[k] += np.real(Cq_coeff[j] * Cq_terms[j][i])

        del (Cq_keys, Cq_terms, Cq_coeff)

        # -----------------------------------------------
        # Non-quadratic K.E. transfer and exchange terms
        # -----------------------------------------------
        inner_prod = lambda a_fft, b_fft: a_fft.conj() * b_fft
        inner_prod2 = lambda a_fft, b_fft: a_fft * b_fft.conj()
        triple_prod_conv = lambda ax_fft, ay_fft, bx_fft, by_fft: (
            inner_prod(ax_fft, self.fnonlinfft_from_uxuy_funcfft(ux, uy, bx_fft))
            + inner_prod(
                ay_fft, self.fnonlinfft_from_uxuy_funcfft(ux, uy, by_fft)
            )
        )
        triple_prod_conv2 = lambda ax_fft, ay_fft, bx_fft, by_fft: (
            inner_prod2(ax_fft, self.fnonlinfft_from_uxuy_funcfft(ux, uy, bx_fft))
            + inner_prod2(
                ay_fft, self.fnonlinfft_from_uxuy_funcfft(ux, uy, by_fft)
            )
        )

        u_udM = triple_prod_conv(ux_fft, uy_fft, Mx_fft, My_fft)
        M_udu = triple_prod_conv2(Mx_fft, My_fft, ux_fft, uy_fft)
        divM = oper.ifft2(oper.divfft_from_vecfft(Mx_fft, My_fft))

        ux_divM = oper.fft2(ux * divM)
        uy_divM = oper.fft2(uy * divM)
        oper.dealiasing(ux_divM, uy_divM)

        u_u_divM = inner_prod(ux_fft, ux_divM) + inner_prod(uy_fft, uy_divM)
        Tnq_fft = 0.5 * np.real(M_udu + u_udM - u_u_divM)
        del (M_udu, u_udM, divM, ux_divM, uy_divM, u_u_divM)

        # --------------------------------------
        # Enstrophy transfer terms
        # --------------------------------------
        Tens_fft = oper.K2 * Tq_fft["GGG"]

        Tq_GGG = self.spectrum2D_from_fft(Tq_fft["GGG"])
        Tq_AGG = self.spectrum2D_from_fft(Tq_fft["AGG"])
        Tq_GAAs = self.spectrum2D_from_fft(Tq_fft["GAAs"])
        Tq_GAAd = self.spectrum2D_from_fft(Tq_fft["GAAd"])
        Tq_AAA = self.spectrum2D_from_fft(Tq_fft["AAA"])
        Tnq = self.spectrum2D_from_fft(Tnq_fft)
        Tens = self.spectrum2D_from_fft(Tens_fft)
        Cq_GG = self.spectrum2D_from_fft(Cq_fft["GG"])
        Cq_AG = self.spectrum2D_from_fft(Cq_fft["AG"])
        Cq_aG = self.spectrum2D_from_fft(Cq_fft["aG"])
        Cq_AA = self.spectrum2D_from_fft(Cq_fft["AA"])

        # Tq_TOT = Tq_GGG + Tq_AGG + Tq_GAAs + Tq_GAAd + Tq_AAA
        # self._checksum_stdout(
        #   GGG=Tq_GGG, GGA=Tq_AGG, AAG=(Tq_GAAs+Tq_GAAd), AAA=Tq_AAA,
        #   TNQ=Tnq, TOTAL=Tq_TOT, debug=True)

        dict_results = {
            "Tq_GGG": Tq_GGG,
            "Tq_AGG": Tq_AGG,
            "Tq_GAAs": Tq_GAAs,
            "Tq_GAAd": Tq_GAAd,
            "Tq_AAA": Tq_AAA,
            "Tnq": Tnq,
            "Cq_GG": Cq_GG,
            "Cq_AG": Cq_AG,
            "Cq_aG": Cq_aG,
            "Cq_AA": Cq_AA,
            "Tens": Tens,
        }

        return dict_results

    def _online_plot_saving(self, dict_results):

        # Tens = dict_results["Tens"]
        Tq_GGG = dict_results["Tq_GGG"]
        Tq_AGG = dict_results["Tq_AGG"]
        Tq_GAAs = dict_results["Tq_GAAs"]
        Tq_GAAd = dict_results["Tq_GAAd"]
        Tq_AAA = dict_results["Tq_AAA"]
        Tq_tot = Tq_GGG + Tq_AGG + Tq_GAAs + Tq_GAAd + Tq_AAA

        Cq_GG = dict_results["Cq_GG"]
        Cq_AG = dict_results["Cq_AG"] + dict_results["Cq_aG"]
        Cq_AA = dict_results["Cq_AA"]
        Cq_tot = Cq_GG + Cq_AG + Cq_AA

        khE = self.oper.khE
        # Piens = cumsum_inv(Tens) * self.oper.deltak
        Pi_tot = cumsum_inv(Tq_tot) * self.oper.deltak
        Pi_GGG = cumsum_inv(Tq_GGG) * self.oper.deltak
        Pi_AGG = cumsum_inv(Tq_AGG) * self.oper.deltak
        Pi_GAAs = cumsum_inv(Tq_GAAs) * self.oper.deltak
        Pi_GAAd = cumsum_inv(Tq_GAAd) * self.oper.deltak
        Pi_AAA = cumsum_inv(Tq_AAA) * self.oper.deltak

        Cflux_tot = cumsum_inv(Cq_tot) * self.oper.deltak
        Cflux_GG = cumsum_inv(Cq_GG) * self.oper.deltak
        Cflux_AG = cumsum_inv(Cq_AG) * self.oper.deltak
        Cflux_AA = cumsum_inv(Cq_AA) * self.oper.deltak

        self.axe_a.plot(
            khE + khE[1], Pi_tot, "k", linewidth=2, label=r"$\Pi_{tot}$"
        )
        self.axe_a.plot(
            khE + khE[1], Pi_GGG, "g--", linewidth=1, label=r"$\Pi_{GGG}$"
        )
        # self.axe_a.plot(
        #     khE+khE[1], Piens, 'g:', linewidth=1, label=r'$\Pi_{ens}$')
        self.axe_a.plot(
            khE + khE[1], Pi_AGG, "m--", linewidth=1, label=r"$\Pi_{GGA}$"
        )
        self.axe_a.plot(
            khE + khE[1], Pi_GAAs, "r:", linewidth=1, label=r"$\Pi_{G\pm\pm}$"
        )
        self.axe_a.plot(
            khE + khE[1], Pi_GAAd, "b:", linewidth=1, label=r"$\Pi_{G\pm\mp}$"
        )
        self.axe_a.plot(
            khE + khE[1], Pi_AAA, "y--", linewidth=1, label=r"$\Pi_{AAA}$"
        )

        self.axe_b.plot(
            khE + khE[1], Cflux_tot, "k", linewidth=2, label=r"$\Sigma C_{tot}$"
        )
        self.axe_b.plot(
            khE + khE[1], Cflux_GG, "g:", linewidth=1, label=r"$\Sigma C_{GG}$"
        )
        self.axe_b.plot(
            khE + khE[1], Cflux_AG, "m--", linewidth=1, label=r"$\Sigma C_{GA}$"
        )
        self.axe_b.plot(
            khE + khE[1], Cflux_AA, "y--", linewidth=1, label=r"$\Sigma C_{AA}$"
        )

        if self.nb_saved_times == 2:
            self.axe_a.set_title(
                "Spectral Energy Budget\n" + self.output.summary_simul
            )
            self.axe_a.legend()
            self.axe_b.legend()
            self.axe_b.set_ylabel(r"$\Sigma C(k_h)$")

    def plot(self, tmin=0, tmax=1000, delta_t=2):

        with h5py.File(self.path_file, "r") as h5file:

            dset_times = h5file["times"]
            dset_khE = h5file["khE"]
            khE = dset_khE[...] + 0.1  # Offset for semilog plots
            times = dset_times[...]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = "plot fluxes 2D" + (
                ", tmin = {0:8.6g} ; tmax = {1:8.6g} ; delta_t = {2:8.6g}"
                + ", imin = {3:8d} ; imax = {4:8d} ; delta_i = {5:8d}"
            ).format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            # -------------------------
            #  Transfer terms
            # -------------------------
            x_left_axe = 0.12
            z_bottom_axe = 0.46
            width_axe = 0.85
            height_axe = 0.47
            size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
            fig1, ax1 = self.output.figure_axe(size_axe=size_axe)
            fig2, ax2 = self.output.figure_axe(size_axe=size_axe)

            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel(r"Transfer fluxes, $\Pi(k_h)$")

            z_bottom_axe = 0.07
            height_axe = 0.27
            size_axe[1] = z_bottom_axe
            size_axe[3] = height_axe
            ax11 = fig1.add_axes(size_axe)
            ax11.set_xlabel("$k_h$")
            ax11.set_ylabel("Transfer terms, $T(k_h)$")

            title = "Spectral Energy Budget\n" + self.output.summary_simul
            ax1.set_title(title)
            ax1.set_xscale("log")
            ax1.axhline()
            ax11.set_title(title)
            ax11.set_xscale("log")
            ax11.axhline()

            P = self.sim.params.forcing.forcing_rate
            norm = 1 if P == 0 else P
            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot, delta_i_plot):
                    transferEtot = 0.0
                    for key in ["GGG", "AGG", "GAAs", "GAAd", "AAA"]:
                        transferEtot += h5file["Tq_" + key][it]
                    PiEtot = cumsum_inv(transferEtot) * self.oper.deltak / norm
                    ax1.plot(khE, PiEtot, "k", linewidth=1)

            Tq_GGG = h5file["Tq_GGG"][imin_plot:imax_plot].mean(0) / norm
            Tq_AGG = h5file["Tq_AGG"][imin_plot:imax_plot].mean(0) / norm
            Tq_GAAs = h5file["Tq_GAAs"][imin_plot:imax_plot].mean(0) / norm
            Tq_GAAd = h5file["Tq_GAAd"][imin_plot:imax_plot].mean(0) / norm
            Tq_AAA = h5file["Tq_AAA"][imin_plot:imax_plot].mean(0) / norm
            Tnq = h5file["Tnq"][imin_plot:imax_plot].mean(0) / norm
            Tens = h5file["Tens"][imin_plot:imax_plot].mean(0) / norm
            Tq_tot = Tq_GGG + Tq_AGG + Tq_GAAs + Tq_GAAd + Tq_AAA

            Pi_GGG = cumsum_inv(Tq_GGG) * self.oper.deltak
            Pi_AGG = cumsum_inv(Tq_AGG) * self.oper.deltak
            Pi_GAAs = cumsum_inv(Tq_GAAs) * self.oper.deltak
            Pi_GAAd = cumsum_inv(Tq_GAAd) * self.oper.deltak
            Pi_AAA = cumsum_inv(Tq_AAA) * self.oper.deltak
            Pi_nq = cumsum_inv(Tnq) * self.oper.deltak
            Pi_ens = cumsum_inv(Tens) * self.oper.deltak
            Pi_tot = Pi_GGG + Pi_AGG + Pi_GAAs + Pi_GAAd + Pi_AAA

            ax1.plot(khE, Pi_GGG, "g--", linewidth=2, label=r"$\Pi_{GGG}$")
            ax1.plot(khE, Pi_AGG, "m--", linewidth=2, label=r"$\Pi_{GGA}$")
            # ax1.plot(khE, Pi_GAAs, 'r:', linewidth=2, label=r'$\Pi_{G\pm\pm}$')
            # ax1.plot(khE, Pi_GAAd, 'b:', linewidth=2, label=r'$\Pi_{G\pm\mp}$')
            ax1.plot(
                khE, Pi_GAAs + Pi_GAAd, "r", linewidth=2, label=r"$\Pi_{GAA}$"
            )
            ax1.plot(khE, Pi_AAA, "y--", linewidth=2, label=r"$\Pi_{AAA}$")
            ax1.plot(khE, Pi_nq, "k--", linewidth=1, label=r"$\Pi^{NQ}$")
            ax1.plot(khE, Pi_tot, "k", linewidth=3, label=r"$\Pi_{tot}$")
            ax1.legend()

            ax11.plot(khE, Tq_GGG, "g--", linewidth=2, label=r"$T_{GGG}$")
            ax11.plot(khE, Tq_AGG, "m--", linewidth=2, label=r"$T_{GGA}$")
            ax11.plot(khE, Tq_GAAs, "r:", linewidth=2, label=r"$T_{G\pm\pm}$")
            ax11.plot(khE, Tq_GAAd, "b:", linewidth=2, label=r"$T_{G\pm\mp}$")
            ax11.plot(khE, Tq_AAA, "y--", linewidth=2, label=r"$T_{AAA}$")
            ax11.plot(khE, Tnq, "k--", linewidth=2, label=r"$T^{NQ}$")
            ax11.plot(khE, Tq_tot, "k", linewidth=3, label=r"$T_{tot}$")
            ax11.legend()
            # -------------------------
            # Quadratic exchange terms
            # -------------------------
            ax2.set_xlabel(r"$k_h$")
            ax2.set_ylabel(r"Exchange fluxes, $\Sigma C$")
            ax2.set_xscale("log")
            ax2.axhline()

            # .. TODO: Normalize with energy??
            exchange_GG = h5file["Cq_GG"][imin_plot:imax_plot].mean(0)
            exchange_AG = h5file["Cq_AG"][imin_plot:imax_plot].mean(0) + h5file[
                "Cq_aG"
            ][imin_plot:imax_plot].mean(0)
            exchange_AA = h5file["Cq_AA"][imin_plot:imax_plot].mean(0)
            exchange_mean = exchange_GG + exchange_AG + exchange_AA

            Cflux_GG = cumsum_inv(exchange_GG) * self.oper.deltak
            Cflux_AG = cumsum_inv(exchange_AG) * self.oper.deltak
            Cflux_AA = cumsum_inv(exchange_AA) * self.oper.deltak
            Cflux_mean = cumsum_inv(exchange_mean) * self.oper.deltak

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot, delta_i_plot):
                    exchangetot = 0.0
                    for key in ["GG", "AG", "aG", "AA"]:
                        exchangetot += h5file["Cq_" + key][it]
                    Cfluxtot = cumsum_inv(exchangetot) * self.oper.deltak
                    ax2.plot(khE, Cfluxtot, "k", linewidth=1)

        ax2.plot(khE, Cflux_mean, "k", linewidth=4, label=r"$\Sigma C_{tot}$")
        ax2.plot(khE, Cflux_GG, "g:", linewidth=2, label=r"$\Sigma C_{GG}$")
        ax2.plot(khE, Cflux_AG, "m--", linewidth=2, label=r"$\Sigma C_{GA}$")
        ax2.plot(khE, Cflux_AA, "y--", linewidth=2, label=r"$\Sigma C_{AA}$")
        ax2.legend()

        ax22 = fig2.add_axes(size_axe)
        ax22.set_xscale("log")
        ax22.axhline()
        ax22.set_xlabel(r"$k_h$")
        ax22.set_ylabel(r"$\Pi_{ens}(k_h)$")

        ax22.plot(khE, Pi_ens, "g", linewidth=3, label=r"$\Pi_{ens}$")
        ax22.legend()

        fig1.canvas.draw()
        fig2.canvas.draw()


class SpectralEnergyBudgetSW1LModified(SpectralEnergyBudgetSW1L):
    def __init__(self, output):
        norm_mode = NormalModeDecompositionModified(output)
        super().__init__(output, norm_mode)

    def _transfer_keys_coeff(self):
        c2 = self.params.c2
        keys = {
            "uuu": ["ux_fft", "urx", "px_ux"],  # Quad. K.E. transfer terms
            "uvu": ["ux_fft", "ury", "py_ux"],
            "vuv": ["uy_fft", "urx", "px_uy"],
            "vvv": ["uy_fft", "ury", "py_uy"],
            "eeu": ["eta_fft", "urx", "px_eta"],  # Quad. A.P.E. transfer terms
            "eev": ["eta_fft", "ury", "py_eta"],
        }

        coeff = {
            "uuu": -1.0,
            "uvu": -1.0,
            "vuv": -1.0,
            "vvv": -1.0,
            "eeu": -c2,
            "eev": -c2,
        }

        return keys, coeff



"""Spectra output (:mod:`fluidsim.solvers.ns2d.strat.output.spectra`)
=====================================================================

.. autoclass:: SpectraNS2DStrat
   :members:
   :private-members:

"""

import h5py

import numpy as np

from fluidsim.base.output.spectra import Spectra


class SpectraNS2DStrat(Spectra):
    """Save and plot spectra."""

    def compute(self):
        """compute the values at one time."""
        # energy_fft = self.output.compute_energy_fft()
        energyK_fft, energyA_fft = self.output.compute_energies_fft()
        energy_fft = energyK_fft + energyA_fft
        energyK_ux_fft, energyK_uy_fft = self.output.compute_energies2_fft()
        energyK, energyA, energyK_ux = self.output.compute_energies()

        # Compute the kinetic energy spectra 1D for the two velocity components
        # and two directions
        spectrum1Dkx_EK_ux, spectrum1Dky_EK_ux = self.spectra1D_from_fft(
            energyK_ux_fft
        )
        spectrum1Dkx_EK_uy, spectrum1Dky_EK_uy = self.spectra1D_from_fft(
            energyK_uy_fft
        )
        spectrum1Dkx_EK, spectrum1Dky_EK = self.spectra1D_from_fft(energyK_fft)

        # Compute the potential energy spectra 1D two directions
        spectrum1Dkx_EA, spectrum1Dky_EA = self.spectra1D_from_fft(energyA_fft)

        # Compute the total energy spectra 1D
        spectrum1Dkx_E, spectrum1Dky_E = self.spectra1D_from_fft(energy_fft)
        # Dictionary with the 1D kinetic energy spectra
        dict_spectra1D = {
            "spectrum1Dkx_EK_ux": spectrum1Dkx_EK_ux,
            "spectrum1Dky_EK_ux": spectrum1Dky_EK_ux,
            "spectrum1Dkx_EK_uy": spectrum1Dkx_EK_uy,
            "spectrum1Dky_EK_uy": spectrum1Dky_EK_uy,
            "spectrum1Dkx_EK": spectrum1Dkx_EK,
            "spectrum1Dky_EK": spectrum1Dky_EK,
            "spectrum1Dkx_EA": spectrum1Dkx_EA,
            "spectrum1Dky_EA": spectrum1Dky_EA,
            "spectrum1Dkx_E": spectrum1Dkx_E,
            "spectrum1Dky_E": spectrum1Dky_E,
        }

        # compute the kinetic energy spectra 2D
        spectrum2D_E = self.spectrum2D_from_fft(energy_fft)
        spectrum2D_EK = self.spectrum2D_from_fft(energyK_fft)
        spectrum2D_EK_ux = self.spectrum2D_from_fft(energyK_ux_fft)
        spectrum2D_EK_uy = self.spectrum2D_from_fft(energyK_uy_fft)
        spectrum2D_EA = self.spectrum2D_from_fft(energyA_fft)
        dict_spectra2D = {
            "spectrum2D_EK_ux": spectrum2D_EK_ux,
            "spectrum2D_EK_uy": spectrum2D_EK_uy,
            "spectrum2D_EK": spectrum2D_EK,
            "spectrum2D_EA": spectrum2D_EA,
            "spectrum2D_E": spectrum2D_E,
        }

        return dict_spectra1D, dict_spectra2D

    def _online_plot_saving(self, dict_spectra1D, dict_spectra2D):
        if (
            self.nx == self.params.oper.ny
            and self.params.oper.Lx == self.params.oper.Ly
        ):
            spectrum2D_EK = dict_spectra2D["spectrum2D_EK"]
            spectrum2D_EA = dict_spectra2D["spectrum2D_EA"]
            spectrum2D = spectrum2D_EK + spectrum2D_EA
            khE = self.oper.khE
            coef_norm = khE ** (3.0)
            self.axe.loglog(khE, spectrum2D * coef_norm, "k")
            lin_inf, lin_sup = self.axe.get_ylim()
            if lin_inf < 10e-6:
                lin_inf = 10e-6
            self.axe.set_ylim([lin_inf, lin_sup])
        else:
            print(
                "you need to implement the ploting "
                "of the spectra for this case"
            )

    def load1d_means(
        self, tmin=0, tmax=None, delta_t=2, versus_kx=True, versus_ky=True
    ):
        means = {}
        with h5py.File(self.path_file1D, "r") as h5file:

            # Open data from file
            dset_times = h5file["times"]
            dset_kxE = h5file["kxE"]
            dset_kyE = h5file["kyE"]
            times = dset_times[...]
            means["times"] = times
            means["kx"] = dset_kxE[...]
            means["ky"] = dset_kyE[...]

            if tmin is None:
                tmin = 0
            if tmax is None:
                tmax = np.max(times)

            # Compute average from tmin and tmax for plot
            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            delta_t = delta_t_save * delta_i_plot
            if delta_i_plot == 0:
                delta_i_plot = 1

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            if versus_kx:
                # Open data set 1D spectra
                dset_spectrum1Dkx_EA = h5file["spectrum1Dkx_EA"]
                dset_spectrum1Dkx_EK = h5file["spectrum1Dkx_EK"]
                dset_spectrum1Dkx_E = h5file["spectrum1Dkx_E"]

                # Average in time between tmin and tmax
                means["E_kx"] = (
                    dset_spectrum1Dkx_E[imin_plot : imax_plot + 1]
                ).mean(0)
                means["EK_kx"] = (
                    dset_spectrum1Dkx_EK[imin_plot : imax_plot + 1]
                ).mean(0)
                means["EA_kx"] = (
                    dset_spectrum1Dkx_EA[imin_plot : imax_plot + 1]
                ).mean(0)

            if versus_ky:
                # Open data set 1D spectra
                dset_spectrum1Dky_EA = h5file["spectrum1Dky_EA"]
                dset_spectrum1Dky_EK = h5file["spectrum1Dky_EK"]
                dset_spectrum1Dky_E = h5file["spectrum1Dky_E"]

                # Average in time between tmin and tmax
                means["E_ky"] = (
                    dset_spectrum1Dky_E[imin_plot : imax_plot + 1]
                ).mean(0)
                means["EK_ky"] = (
                    dset_spectrum1Dky_EK[imin_plot : imax_plot + 1]
                ).mean(0)
                means["EA_ky"] = (
                    dset_spectrum1Dky_EA[imin_plot : imax_plot + 1]
                ).mean(0)

        print(
            """load spectra
        tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
        imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
        )

        return means

    def plot1d(
        self,
        tmin=0,
        tmax=None,
        delta_t=2,
        coef_compensate=5 / 3,
        level2=1.0,
        level3=1.0,
        yrange=5,
    ):
        """Plot spectra one-dimensional."""
        print(
            "plot1d(tmin={}, tmax={}, delta_t={:.2f},".format(tmin, tmax, delta_t)
            + f" coef_compensate={coef_compensate:.3f})"
        )

        means = self.load1d_means(
            tmin, tmax, delta_t, versus_kx=True, versus_ky=True
        )

        kx = means["kx"]
        ky = means["ky"]

        # Parameters figure E(k_x)
        fig, ax = self.output.figure_axe()
        ax.set_xlabel("$k_x$, $k_z$")
        ax.set_ylabel(r"$E(k)k^{{{}}}$".format(round(coef_compensate, 2)))
        ax.set_title("1D spectra\n" + self.output.summary_simul)
        ax.set_xscale("log")
        ax.set_yscale("log")
        # ax.set_ylim(ymin=1e-6, ymax=1e3)

        id_kx_dealiasing = (
            np.argmin(abs(kx - (kx.max() * self.sim.oper.coef_dealiasing))) - 1
        )
        id_ky_dealiasing = (
            np.argmin(abs(ky - (ky.max() * self.sim.oper.coef_dealiasing))) - 1
        )

        # Remove modes dealiased.
        E_kx_plot = means["E_kx"][:id_kx_dealiasing]
        EK_kx_plot = means["EK_kx"][:id_kx_dealiasing]
        EA_kx_plot = means["EA_kx"][:id_kx_dealiasing]
        kx_plot = kx[:id_kx_dealiasing]

        # Remove shear modes if there is no energy on them.
        if self.sim.params.oper.NO_SHEAR_MODES:
            E_kx_plot = E_kx_plot[1:]
            EK_kx_plot = EK_kx_plot[1:]
            EA_kx_plot = EA_kx_plot[1:]
            kx_plot = kx_plot[1:]

        # Compute k_b: L_b = U / N
        U = np.sqrt(np.mean(abs(self.sim.state.get_var("ux")) ** 2))
        k_b = self.sim.params.N / U
        ax.axvline(x=k_b, color="y", linestyle="--", label="$k_b$")

        # Plot ozmidov scale
        k_o = (self.sim.params.N ** 3 / self.sim.params.forcing.forcing_rate) ** (
            1 / 2
        )
        ax.axvline(x=k_o, color="y", linestyle=":", label="$k_o$")

        # ax.plot(
        # kx_plot, E_kx_plot * kx_plot ** coef_compensate, "k", label="$E(k_x)$"
        # )
        ax.plot(
            kx_plot,
            EK_kx_plot * kx_plot ** coef_compensate,
            "r",
            label="$E_K(k_x)$",
        )
        ax.plot(
            kx_plot,
            EA_kx_plot * kx_plot ** coef_compensate,
            "b",
            label="$E_A(k_x)$",
        )

        # Remove modes dealiased.
        E_ky_plot = means["E_ky"][:id_ky_dealiasing]
        EK_ky_plot = means["EK_ky"][:id_ky_dealiasing]
        EA_ky_plot = means["EA_ky"][:id_ky_dealiasing]
        ky_plot = ky[:id_ky_dealiasing]

        # Remove shear modes if there is no energy on them.
        if self.sim.params.oper.NO_SHEAR_MODES:
            E_ky_plot = E_ky_plot[1:]
            EK_ky_plot = EK_ky_plot[1:]
            EA_ky_plot = EA_ky_plot[1:]
            ky_plot = ky_plot[1:]

        # ax.plot(
        # ky_plot,
        # E_ky_plot * ky_plot ** coef_compensate,
        # "k--",
        # label="$E(k_z)$",
        # )
        ax.plot(
            ky_plot,
            EK_ky_plot * ky_plot ** coef_compensate,
            "r--",
            label="$E_K(k_z)$",
        )
        ax.plot(
            ky_plot,
            EA_ky_plot * ky_plot ** coef_compensate,
            "b--",
            label="$E_A(k_z)$",
        )

        # Plot scaling lines
        kx = kx_plot[40:id_kx_dealiasing]
        if level3:
            ax.plot(
                kx,
                level3 * kx ** (-3) * kx ** coef_compensate,
                "k:",
                label=r"$k^{-3}$",
            )
        if level2:
            ax.plot(
                kx,
                level2 * kx ** (-2) * kx ** coef_compensate,
                "k-.",
                label=r"$k^{-2}$",
            )

        # Plot forcing wave-number k_f
        nkmax = self.sim.params.forcing.nkmax_forcing
        nkmin = self.sim.params.forcing.nkmin_forcing
        pforcing = self.sim.params.forcing
        if pforcing.enable and pforcing.type == "tcrandom_anisotropic":
            angle = pforcing.tcrandom_anisotropic.angle
            try:
                if angle.endswith("°"):
                    angle = np.pi / 180 * float(angle[:-1])
            except AttributeError:
                pass

            # Band forcing region kx
            k_fxmin = nkmin * self.sim.oper.deltak * np.sin(angle)
            k_fxmin = max(k_fxmin, self.sim.oper.deltakx)
            k_fxmax = nkmax * self.sim.oper.deltak * np.sin(angle)

            # Band forcing region ky
            k_fymin = nkmin * self.sim.oper.deltak * np.cos(angle)
            k_fymin = max(k_fymin, self.sim.oper.deltaky)
            k_fymax = nkmax * self.sim.oper.deltak * np.cos(angle)

            # Plot forcing band
            ax.axvspan(k_fxmin, k_fxmax, alpha=0.15, color="black")
            ax.axvspan(k_fymin, k_fymax, alpha=0.15, color="black")

        # Set limits axis y
        if yrange is not None:
            ymax = 2 * max(
                (E_kx_plot * kx_plot ** coef_compensate).max(),
                (E_ky_plot * ky_plot ** coef_compensate).max(),
            )
            ax.set_ylim((10 ** (-yrange) * ymax, ymax))

        ax.legend()

        # from ipdb import set_trace
        # set_trace()

    def load2d_means(self, tmin=0, tmax=1000, delta_t=2):

        means = {}
        # Load data from file
        with h5py.File(self.path_file2D, "r") as h5file:
            dset_times = h5file["times"]
            means["kh"] = h5file["khE"][...]
            times = dset_times[...]
            means["times"] = times

            # Compute average from tmin and tmax for plot
            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            print(
                """plot 2D spectra
            tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
            imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            dset_spectrum_E = h5file["spectrum2D_E"]
            dset_spectrum_EK = h5file["spectrum2D_EK"]
            dset_spectrum_EA = h5file["spectrum2D_EA"]

            means["E"] = dset_spectrum_E[imin_plot : imax_plot + 1].mean(0)
            means["EK"] = dset_spectrum_EK[imin_plot : imax_plot + 1].mean(0)
            means["EA"] = dset_spectrum_EA[imin_plot : imax_plot + 1].mean(0)

        return means

    def plot2d(self, tmin=0, tmax=1000, delta_t=2, coef_compensate=3):
        """Plot 2D spectra."""

        print(
            "plot2s(tmin={}, tmax={}, delta_t={:.2f},".format(tmin, tmax, delta_t)
            + f" coef_compensate={coef_compensate:.3f})"
        )

        means = self.load2d_means(tmin, tmax, delta_t)

        kh = means["kh"]
        E = means["E"]
        EK = means["EK"]
        EA = means["EA"]

        # Parameters figure
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(r"$k$")
        ax.set_ylabel(r"$E(k)$")
        ax.set_title("2D spectra\n" + self.output.summary_simul)

        ax.set_xscale("log")
        ax.set_yscale("log")

        ax.plot(kh, E, "k", label="E")
        ax.plot(kh, EK, "r", label="EK")
        ax.plot(kh, EA, "b", label="EA")

        ax.legend()



"""Time stepping (:mod:`fluidsim.base.time_stepping.finite_diff`)
=======================================================================

Provides:

.. autoclass:: TimeSteppingFiniteDiffCrankNicolson
   :members:
   :private-members:

"""

from copy import deepcopy

import numpy as np
import scipy.sparse as sparse
from scipy.sparse.linalg import spsolve

from fluidsim.base.setofvariables import SetOfVariables

from .base import TimeSteppingBase


class TimeSteppingFiniteDiffCrankNicolson(TimeSteppingBase):
    """Time stepping class for finite-difference solvers."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        TimeSteppingBase._complete_params_with_default(params)
        params.time_stepping.type_time_scheme = "RK2"

    def __init__(self, sim):
        super().__init__(sim)

        self._init_compute_time_step()
        self._init_time_scheme()

        self.L = sim.linear_operator()

    def one_time_step_computation(self):
        """One time step"""
        self._time_step_RK()
        if np.isnan(np.min(self.sim.state.state_phys)):
            raise ValueError(f"nan at it = {self.it}, t = {self.t:.4f}")

    def _time_step_RK2(self):
        r"""Advance in time the variables with the Runge-Kutta 2 method.

        .. _rk2timeschemeFiniteDiff:

        Notes
        -----

        .. Look at Simson KTH documentation...
           (http://www.mech.kth.se/~mattias/simson-user-guide-v4.0.pdf)

        The Runge-Kutta 2 method computes an approximation of the
        solution after a time increment :math:`dt`. We denote the
        initial time :math:`t = 0`.

        For the finite difference schemes, We consider an equation of the form

        .. math:: \p_t S = L S + N(S),

        The linear term can be treated with an implicit method while
        the nonlinear term have to be treated with an explicit method
        (see for example `Explicit and implicit methods
        <http://en.wikipedia.org/wiki/Explicit_and_implicit_methods>`_).

        - Approximation 1:

          For the first step where the nonlinear term is approximated
          as :math:`N(S) \simeq N(S_0)`, we obtain

          .. math::
             \left( 1 - \frac{dt}{4} L \right) S_{A1dt/2}
             \simeq \left( 1 + \frac{dt}{4} L \right) S_0 + N(S_0)dt/2

          Once the right-hand side has been computed, a linear
          equation has to be solved. It is not efficient to invert the
          matrix :math:`1 + \frac{dt}{2} L` so other methods have to
          be used, as the `Thomas algorithm
          <http://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm>`_,
          or algorithms based on the LU or the QR decompositions.

        - Approximation 2:

            The nonlinear term is then approximated as :math:`N(S)
            \simeq N(S_{A1dt/2})`, which gives

            .. math::
               \left( 1 - \frac{dt}{2} L \right) S_{A2dt}
               \simeq \left( 1 + \frac{dt}{2} L \right) S_0 + N(S_{A1dt/2})dt

        """
        dt = self.deltat
        sim = self.sim
        identity = sparse.identity(sim.state.state_phys.size)

        # it seems that there is a bug with the proper RK2 method
        # (it "goes too fast")

        # # approximation 1 (at t + dt/2 -> "A1dt2"):
        # tendenciesNL_0 = sim.tendencies_nonlin()
        # rhs_A1dt2 = self.right_hand_side(sim.state.state_phys,
        #                                  tendenciesNL_0, dt/2)

        # A_A1dt2 = identity - dt/4*self.L
        # S_A1dt2 = self.invert_to_get_solution(A_A1dt2, rhs_A1dt2)
        # del(rhs_A1dt2, A_A1dt2)

        # # approximation 2 (at t + dt -> "A2dt"):
        # tendenciesNL_1 = sim.tendencies_nonlin(S_A1dt2)
        # rhs_A2dt = self.right_hand_side(S_A1dt2, tendenciesNL_1, dt)
        # A_A2dt = identity - dt/2*self.L
        # sim.state.state_phys = deepcopy(
        #     self.invert_to_get_solution(A_A2dt, rhs_A2dt))

        # it seems to work with the basic Newton time stepping:
        tendenciesNL_0 = sim.tendencies_nonlin()
        rhs_A1dt = self.right_hand_side(sim.state.state_phys, tendenciesNL_0, dt)
        A_A1dt = identity - dt / 2 * self.L
        sim.state.state_phys = deepcopy(
            self.invert_to_get_solution(A_A1dt, rhs_A1dt)
        )

    def right_hand_side(self, S, N, dt):
        return S.ravel() + dt / 2 * self.L.dot(S.flat) + dt * N.ravel()

    def invert_to_get_solution(self, A, b):
        """Solve the linear system :math:`Ax = b`."""
        state_phys = self.sim.state.state_phys
        arr = spsolve(A, b).reshape(state_phys.shape)
        return SetOfVariables(
            input_array=arr, keys=state_phys.keys, info=state_phys.info
        )



"""One-layer shallow-water (Saint Venant) equations solver (:mod:`fluidsim.solvers.sw1l.solver`)
================================================================================================

This module provides two classes defining the pseudo-spectral solver for
biperiodic one-layer shallow water equations, expressed using primitive
variables, i.e. the horizontal velocities and surface dispacement.

.. autoclass:: InfoSolverSW1L
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:



"""

from transonic import jit, Array
from fluiddyn.util import mpi

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)

A = Array[float, "2d"]


@jit
def compute_Frot(rot: A, ux: A, uy: A, f: float):
    """Compute cross-product of absolute potential vorticity with velocity."""
    if f != 0:
        rot_abs = rot + f
    else:
        rot_abs = rot

    F1x = rot_abs * uy
    F1y = -rot_abs * ux

    return F1x, F1y


@jit
def compute_pressure(c2: float, eta: A, ux: A, uy: A):
    return c2 * eta + 0.5 * (ux ** 2 + uy ** 2)


class InfoSolverSW1L(InfoSolverPseudoSpectral):
    """Information about the solver SW1L."""

    def _init_root(self):
        """The simulation object is instantiated with classes defined in this
        function.

        The super function `InfoSolverPseudoSpectral._init_root` is
        called first. A few classes defined by this function are retained as it
        is:

        - :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`

        - :class:`fluidsim.solvers.sw1l.OperatorsPseudoSpectralSW1L`

        - :class:`fluidsim.base.preprocess.pseudo_spect.PreprocessPseudoSpectral`

        Solver-specific first-level classes are added:

        - :class:`fluidsim.solvers.sw1l.solver.Simul`

        - :class:`fluidsim.solvers.sw1l.state.StateSW1L`

        - :class:`fluidsim.solvers.sw1l.init_fields.InitFieldsSW1L`

        - :class:`fluidsim.solvers.sw1l.output.OutputSW1L`

        - :class:`fluidsim.solvers.sw1l.forcing.ForcingSW1L`

        """
        super()._init_root()

        package = "fluidsim.solvers.sw1l"

        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "SW1L"

        classes = self.classes

        classes.Operators.module_name = package + ".operators"
        classes.Operators.class_name = "OperatorsPseudoSpectralSW1L"

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateSW1L"

        classes.InitFields.module_name = package + ".init_fields"
        classes.InitFields.class_name = "InitFieldsSW1L"

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "OutputSW1L"

        classes.Forcing.module_name = package + ".forcing"
        classes.Forcing.class_name = "ForcingSW1L"


class Simul(SimulBasePseudoSpectral):
    """A solver of the shallow-water 1 layer equations (SW1L).

    .. inheritance-diagram:: Simul

    """

    InfoSolver = InfoSolverSW1L

    @staticmethod
    def _complete_params_with_default(params):
        r"""This static method is used to complete the *params* container.

        Notes
        -----

        - :math:`f` is the system rotation.

        - :math:`c` is the phase speed of surface gravity waves in the short-
          wave limit.

        - :math:`k_d` is the Rossby deformation wavenumber.

        - :math:`\beta` is differential rotation parameter, approximately given
          by :math:`\partial_y f`.

        The present solver will not work in the beta-plane. However a
        vorticity-divergence formulation can be solved with a beta term.

        """
        SimulBasePseudoSpectral._complete_params_with_default(params)

        attribs = {"f": 0, "c2": 20, "kd2": 0, "beta": 0}
        params._set_attribs(attribs)

    def __init__(self, params):
        # Parameter(s) specific to this solver

        params.f = float(params.f)
        params.c2 = float(params.c2)

        params.kd2 = params.f ** 2 / params.c2
        if params.beta != 0:
            raise NotImplementedError(
                "Do not use this solver for beta-plane! "
                "Equations are non-periodic in this formulation."
            )

        super().__init__(params)

        if mpi.rank == 0:
            self.output.print_stdout(
                "c2 = {:6.5g} ; f = {:6.5g} ; kd2 = {:6.5g}".format(
                    params.c2, params.f, params.kd2
                )
            )

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the horizontal velocities and
            surface displacement scalar in Fourier space.  When `state_spect` is
            provided, the variables vorticity and the velocities and surface
            displacement are computed from it, otherwise, they are taken from
            the global state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies_fft : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the velocities
            :math:`\mathbf u` and the surface displacement scalar :math:`\eta`.

        Notes
        -----

        .. |p| mathmacro:: \partial

        The one-layer shallow water equations can be written as:


        .. math:: \partial_t \mathbf u  = - \nabla (|\mathbf u|^2/2 + c^2 \eta)
                        - (\zeta + f) \times \mathbf u
        .. math:: \partial_t \eta = - \nabla. ((\eta + H) \mathbf u)

        This function computes all terms on the RHS of the equations above,
        including the nonlinear term. The linear dissipation term (not shown
        above) is computed implicitly, as demonstrated in
        :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`.

        """
        oper = self.oper

        if state_spect is None:
            state_phys = self.state.state_phys
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        F1x, F1y = compute_Frot(rot, ux, uy, self.params.f)
        gradx_fft, grady_fft = oper.gradfft_from_fft(
            oper.fft2(compute_pressure(self.params.c2, eta, ux, uy))
        )
        oper.dealiasing(gradx_fft, grady_fft)

        Fx_fft = tendencies_fft.get_var("ux_fft")
        Fy_fft = tendencies_fft.get_var("uy_fft")
        Feta_fft = tendencies_fft.get_var("eta_fft")

        Fx_fft[:] = oper.fft2(F1x) - gradx_fft
        Fy_fft[:] = oper.fft2(F1y) - grady_fft

        Feta_fft[:] = -oper.divfft_from_vecfft(
            oper.fft2((eta + 1) * ux), oper.fft2((eta + 1) * uy)
        )

        oper.dealiasing(tendencies_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft


if __name__ == "__main__":

    import numpy as np

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 32
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 1.0
    # params.time_stepping.USE_CFL = False
    # params.time_stepping.deltat0 = 0.01

    params.init_fields.type = "noise"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 0.5
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = True

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "eta"

    sim = Simul(params)

    sim.output.phys_fields.plot(key_field="eta")
    sim.time_stepping.start()
    sim.output.phys_fields.plot(key_field="eta")

    fld.show()



"""Spatial means output (:mod:`fluidsim.solvers.ns2d.output.spatial_means`)
===========================================================================

.. autoclass:: SpatialMeansNS2D
   :members:
   :private-members:

"""

import os
import numpy as np
import matplotlib.pyplot as plt


from fluiddyn.util import mpi

from fluidsim.base.output.spatial_means import SpatialMeansBase


class SpatialMeansNS2D(SpatialMeansBase):
    """Spatial means output."""

    def _save_one_time(self):
        tsim = self.sim.time_stepping.t
        self.t_last_save = tsim

        energy_fft = self.output.compute_energy_fft()
        enstrophy_fft = self.output.compute_enstrophy_fft()
        energy = self.sum_wavenumbers(energy_fft)
        enstrophy = self.sum_wavenumbers(enstrophy_fft)

        f_d, f_d_hypo = self.sim.compute_freq_diss()
        epsK = self.sum_wavenumbers(f_d * 2 * energy_fft)
        epsK_hypo = self.sum_wavenumbers(f_d_hypo * 2 * energy_fft)
        epsZ = self.sum_wavenumbers(f_d * 2 * enstrophy_fft)
        epsZ_hypo = self.sum_wavenumbers(f_d_hypo * 2 * enstrophy_fft)

        if self.sim.params.forcing.enable:
            deltat = self.sim.time_stepping.deltat
            Frot_fft = self.sim.forcing.get_forcing().get_var("rot_fft")
            Fx_fft, Fy_fft = self.vecfft_from_rotfft(Frot_fft)

            rot_fft = self.sim.state.state_spect.get_var("rot_fft")
            ux_fft, uy_fft = self.vecfft_from_rotfft(rot_fft)

            PZ1_fft = np.real(rot_fft.conj() * Frot_fft)
            PZ2_fft = abs(Frot_fft) ** 2

            PZ1 = self.sum_wavenumbers(PZ1_fft)
            PZ2 = deltat / 2 * self.sum_wavenumbers(PZ2_fft)

            PK1_fft = (
                np.real(
                    ux_fft.conj() * Fx_fft
                    + ux_fft * Fx_fft.conj()
                    + uy_fft.conj() * Fy_fft
                    + uy_fft * Fy_fft.conj()
                )
                / 2
            )
            PK2_fft = (abs(Fx_fft) ** 2 + abs(Fy_fft) ** 2) * deltat / 2

            PK1 = self.sum_wavenumbers(PK1_fft)
            PK2 = self.sum_wavenumbers(PK2_fft)

        if mpi.rank == 0:
            epsK_tot = epsK + epsK_hypo

            self.file.write(f"####\ntime = {tsim:11.5e}\n")
            to_print = (
                "E    = {:11.5e} ; Z         = {:11.5e} \n"
                "epsK = {:11.5e} ; epsK_hypo = {:11.5e} ; epsK_tot = {:11.5e} \n"
                "epsZ = {:11.5e} ; epsZ_hypo = {:11.5e} ; epsZ_tot = {:11.5e} \n"
            ).format(
                energy,
                enstrophy,
                epsK,
                epsK_hypo,
                epsK + epsK_hypo,
                epsZ,
                epsZ_hypo,
                epsZ + epsZ_hypo,
            )
            self.file.write(to_print)

            if self.sim.params.forcing.enable:
                PK_tot = PK1 + PK2
                to_print = (
                    "PK1  = {:11.5e} ; PK2       = {:11.5e} ; PK_tot   = {:11.5e} \n"
                    "PZ1  = {:11.5e} ; PZ2       = {:11.5e} ; PZ_tot   = {:11.5e} \n"
                ).format(PK1, PK2, PK1 + PK2, PZ1, PZ2, PZ1 + PZ2)
                self.file.write(to_print)

            self.file.flush()
            os.fsync(self.file.fileno())

        if self.has_to_plot and mpi.rank == 0:

            self.axe_a.plot(tsim, energy, "k.")

            self.axe_b.plot(tsim, epsK_tot, "k.")
            if self.sim.params.forcing.enable:
                self.axe_b.plot(tsim, PK_tot, "m.")

            if tsim - self.t_last_show >= self.period_show:
                self.t_last_show = tsim
                fig = self.axe_a.get_figure()
                fig.canvas.draw()

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.path_file) as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_PK = []
        lines_PZ = []
        lines_epsK = []
        lines_epsZ = []

        for il, line in enumerate(lines):
            if line.startswith("time ="):
                lines_t.append(line)
            if line.startswith("E    ="):
                lines_E.append(line)
            if line.startswith("PK1  ="):
                lines_PK.append(line)
            if line.startswith("PZ1  ="):
                lines_PZ.append(line)
            if line.startswith("epsK ="):
                lines_epsK.append(line)
            if line.startswith("epsZ ="):
                lines_epsZ.append(line)

        nt = len(lines_t)
        # pa: why ?
        # if nt > 1:
        #     nt -= 1

        t = np.empty(nt)
        E = np.empty(nt)
        Z = np.empty(nt)
        if self.sim.params.forcing.enable:
            PK1 = np.empty(nt)
            PK2 = np.empty(nt)
            PK_tot = np.empty(nt)
            PZ1 = np.empty(nt)
            PZ2 = np.empty(nt)
            PZ_tot = np.empty(nt)
        epsK = np.empty(nt)
        epsK_hypo = np.empty(nt)
        epsK_tot = np.empty(nt)
        epsZ = np.empty(nt)
        epsZ_hypo = np.empty(nt)
        epsZ_tot = np.empty(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            t[il] = float(words[2])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            Z[il] = float(words[6])

            if self.sim.params.forcing.enable:
                line = lines_PK[il]
                words = line.split()
                PK1[il] = float(words[2])
                PK2[il] = float(words[6])
                PK_tot[il] = float(words[10])

                line = lines_PZ[il]
                words = line.split()
                PZ1[il] = float(words[2])
                PZ2[il] = float(words[6])
                PZ_tot[il] = float(words[10])

            line = lines_epsK[il]
            words = line.split()
            epsK[il] = float(words[2])
            epsK_hypo[il] = float(words[6])
            epsK_tot[il] = float(words[10])

            line = lines_epsZ[il]
            words = line.split()
            epsZ[il] = float(words[2])
            epsZ_hypo[il] = float(words[6])
            epsZ_tot[il] = float(words[10])

        dict_results["t"] = t
        dict_results["E"] = E
        dict_results["Z"] = Z

        if self.sim.params.forcing.enable:
            dict_results["PK1"] = PK1
            dict_results["PK2"] = PK2
            dict_results["PK_tot"] = PK_tot

            dict_results["PZ1"] = PZ1
            dict_results["PZ2"] = PZ2
            dict_results["PZ_tot"] = PZ_tot

        dict_results["epsK"] = epsK
        dict_results["epsK_hypo"] = epsK_hypo
        dict_results["epsK_tot"] = epsK_tot

        dict_results["epsZ"] = epsZ
        dict_results["epsZ_hypo"] = epsZ_hypo
        dict_results["epsZ_tot"] = epsZ_tot
        return dict_results

    def plot_dt_energy(self):
        """
        Checks if dE/dt = energy_injection - energy_dissipation.

        """
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]

        epsK_tot = dict_results["epsK_tot"]

        if self.sim.params.forcing.enable:
            PK_tot = dict_results["PK_tot"]
            model = PK_tot - epsK_tot
        else:
            model = -epsK_tot

        dtE = np.gradient(E, t)

        fig, ax = plt.subplots()
        ax.set_xlabel("t")
        ax.plot(t, dtE, label="dE/dt")

        ax.plot(t, model, label=r"$P_E - \epsilon$")

        ax.legend()

        fig.tight_layout()

    def plot_dt_enstrophy(self):
        """
        Checks dZ/dt = enstrophy_injection - enstrophy_dissipation.
        """
        dict_results = self.load()

        t = dict_results["t"]
        Z = dict_results["Z"]

        epsZ_tot = dict_results["epsZ_tot"]

        if self.sim.params.forcing.enable:
            PZ_tot = dict_results["PZ_tot"]
            PZ1 = dict_results["PZ1"]
            PZ2 = dict_results["PZ2"]
            model = PZ_tot - epsZ_tot
        else:
            model = -epsZ_tot

        dtZ = np.gradient(Z, t)

        fig, axes = plt.subplots(2)
        ax = axes[0]
        ax.plot(t, dtZ, label="dZ/dt")
        ax.plot(t, model, label=r"$P_Z-\epsilon_Z$")
        ax.legend()

        ax = axes[1]
        ax.plot(t, epsZ_tot, label=r"$\epsilon_{Z}$")
        ax.plot(t, PZ_tot, label=r"$P_{Z}$")
        ax.plot(t, PZ1, label=r"$P_{Z1}$")
        ax.plot(t, PZ2, label=r"$P_{Z2}$")
        ax.legend()

        fig.tight_layout()

    def plot(self):
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        Z = dict_results["Z"]

        epsK_hypo = dict_results["epsK_hypo"]
        epsK_tot = dict_results["epsK_tot"]

        epsZ_hypo = dict_results["epsZ_hypo"]
        epsZ_tot = dict_results["epsZ_tot"]

        fig, axes = plt.subplots(2)
        fig.suptitle("Energy and enstrophy")

        ax0 = axes[0]
        ax0.set_ylabel("$E(t)$")
        ax0.plot(t, E, "k.-", linewidth=2, label=r"$E$")
        ax0.legend()

        ax1 = axes[1]
        ax1.set_ylabel("$Z(t)$")
        ax1.set_xlabel("$t$")
        ax1.plot(t, Z, "k", linewidth=2, label=r"$Z$")
        ax1.legend()

        fig, axes = plt.subplots(2)
        fig.suptitle("Dissipation of energy and enstrophy")

        ax0 = axes[0]

        ax0.set_ylabel(r"$\epsilon_K(t)$")

        ax0.plot(t, epsK_tot, "k", linewidth=2, label=r"$\epsilon$")

        ax1 = axes[1]
        ax1.set_xlabel("$t$")
        ax1.set_ylabel(r"$\epsilon_Z(t)$")
        ax1.plot(t, epsZ_tot, "k", linewidth=2, label=r"$\epsilon_Z$")

        if self.sim.params.forcing.enable:
            PK1 = dict_results["PK1"]
            PK2 = dict_results["PK2"]
            PK_tot = dict_results["PK_tot"]
            PZ_tot = dict_results["PZ_tot"]
            ax0.plot(t, PK_tot, "c", linewidth=2, label="$P_K$")
            ax0.plot(t, PK1, "c--", linewidth=1, label="$P_{K1}$")
            ax0.plot(t, PK2, "c:", linewidth=1, label="$P_{K2}$")
            ax1.plot(t, PZ_tot, "c", linewidth=2, label="$P_Z$")
            ax0.set_ylabel("P_E(t), epsK(t)")
            ax1.set_ylabel("P_Z(t), epsZ(t)")

        if self.sim.params.nu_m4 != 0:
            ax0.plot(t, epsK_hypo, "g", linewidth=1, label=r"$\epsilon_{hypo}$")
            ax1.plot(t, epsZ_hypo, "g", linewidth=1, label=r"$\epsilon_{Zhypo}$")

        ax0.legend()
        ax1.legend()



"""State of the variables (:mod:`fluidsim.base.state`)
============================================================

Provides:

.. autoclass:: StateBase
   :members:
   :private-members:

.. autoclass:: StatePseudoSpectral
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables


class StateBase:
    """Contains the state variables and handles the access to fields.

    Parameters
    ----------

    sim : child of :class:`fluidsim.base.solvers.base.SimulBase`

    oper : Optional[operators]

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_phys": ["X"],
                "keys_computable": [],
                "keys_phys_needed": ["X"],
            }
        )

    def __init__(self, sim, oper=None):
        self.sim = sim
        self.params = sim.params
        if oper is None:
            self.oper = sim.oper
        else:
            self.oper = oper

        # creation of the SetOfVariables state_spect and state_phys
        self.keys_state_phys = sim.info.solver.classes.State.keys_state_phys

        try:
            self.keys_computable = sim.info.solver.classes.State.keys_computable
        except AttributeError:
            self.keys_computable = []

        self.state_phys = SetOfVariables(
            keys=self.keys_state_phys,
            shape_variable=self.oper.shapeX_loc,
            dtype=np.float64,
            info="state_phys",
        )
        self.vars_computed = {}
        self.it_computed = {}

        self.is_initialized = False

    def compute(self, key):
        """Compute a not stored variable from the stored variables"""
        raise ValueError('No method to compute key "' + key + '"')

    def clear_computed(self):
        """Clear the stored computed variables."""
        self.vars_computed.clear()

    def has_vars(self, *keys):
        """Checks if all of the keys are present in the union of
        ``keys_state_phys`` and ``keys_computable``.

        Parameters
        ----------
        keys: str, str ...
            Strings indicating state variable names.

        Returns
        -------
        bool

        Examples
        --------
        >>> sim.state.has_vars('ux', 'uy')
        >>> sim.state.has_vars('ux')
        >>> sim.state.has_vars('ux', 'vx', strict=False)

        .. todo::

           ``strict=True`` can be a Python 3 compatible keywords-only argument
           with the function like::

               def has_vars(self, *keys, strict=True):
                   ...
                   if strict:
                       return keys.issubset(keys_state)
                   else:
                       return len(keys.intersection(keys_state)) > 0

            When ``True``, checks if all keys form a subset of state keys. When
            ``False``, checks if the intersection of the keys and the state keys
            has atleast one member.

        """
        keys_state = set(self.keys_state_phys + self.keys_computable)
        keys = set(keys)
        return keys.issubset(keys_state)

    def get_var(self, key):
        """Get a physical variable (from the storage array or computed).

        This is one of the main method of the state classes.

        It tries to return the array corresponding to a physical variable. If
        it is stored in the main storage array of the state class, it is
        directly returned.  Otherwise, we try to compute the quantity with the
        method :func:`compute`.

        It should not be necessary to redefine this method in child class.

        """
        if key in self.keys_state_phys:
            return self.state_phys.get_var(key)

        else:
            it = self.sim.time_stepping.it
            if key in self.vars_computed and it == self.it_computed[key]:
                return self.vars_computed[key]

            else:
                value = self.compute(key)
                self.vars_computed[key] = value
                self.it_computed[key] = it
                return value

    def __call__(self, key):
        raise DeprecationWarning(
            "Do not call a state object. " "Instead, use get_var method."
        )

    def __setitem__(self, key, value):
        """General setter function to set the value of a variable

        It should not be necessary to redefine this method in child class.
        """
        if key in self.keys_state_phys:
            self.state_phys.set_var(key, value)
        else:
            raise ValueError('key "' + key + '" is not known')

    def can_this_key_be_obtained(self, key):
        """To check whether a variable can be obtained.

        .. deprecated:: 0.2.0
           Use ``has_vars`` method instead.

        """
        raise DeprecationWarning(
            "Do not call can_this_key_be_obtained. "
            "Instead, use has_vars method."
        )

    def init_statephys_from(self, **kwargs):
        """Initialize `state_phys` from arrays.

        Parameters
        ----------

        **kwargs : {key: array, ...}

          keys and arrays used for the initialization. The other keys
          are set to zero.

        Examples
        --------

        .. code-block:: python

           kwargs = {'a': Fa}
           init_statespect_from(**kwargs)

           init_statespect_from(ux=ux, uy=uy, eta=eta)

        """
        self.state_phys[:] = 0.0

        for key, value in list(kwargs.items()):
            if key not in self.keys_state_phys:
                raise ValueError(
                    f'Do not know how to initialize with key "{key}".'
                )

            self.state_phys.set_var(key, value)


class StatePseudoSpectral(StateBase):
    """Contains the state variables and handles the access to fields.

    This is the general class for the pseudo-spectral solvers.

    Parameters
    ----------

    sim : child of :class:`fluidsim.base.solvers.base.SimulBase`

    oper : Optional[operators]

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """

        StateBase._complete_info_solver(info_solver)

        info_solver.classes.State.keys_state_phys = ["ux", "uy"]
        info_solver.classes.State.keys_phys_needed = ["ux", "uy"]

        info_solver.classes.State._set_attribs(
            {"keys_state_spect": ["ux_fft", "uy_fft"]}
        )

    def __init__(self, sim, oper=None):

        super().__init__(sim, oper)

        self.keys_state_spect = sim.info.solver.classes.State.keys_state_spect
        self.state_spect = SetOfVariables(
            keys=self.keys_state_spect,
            shape_variable=self.oper.shapeK_loc,
            dtype=np.complex128,
            info="state_spect",
        )

    def has_vars(self, *keys):
        """Checks if all of the keys are present in the union of
        ``keys_state_phys``, ``keys_computable``, and ``keys_state_spect``.

        Parameters
        ----------
        keys: str, str ...
            Strings indicating state variable names.

        Returns
        -------
        bool

        Examples
        --------
        >>> sim.state.has_vars('ux', 'uy', 'ux_fft')
        >>> sim.state.has_vars('rot')

        """
        keys_state = set(
            self.keys_state_phys + self.keys_computable + self.keys_state_spect
        )
        keys = set(keys)
        return keys.issubset(keys_state)

    def get_var(self, key):
        """Get a variable (from the storage arrays or computed).

        This is one of the main method of the state classes.

        It tries to return the array corresponding to a physical variable. If
        it is stored in the main storage arrays (`state_phys` and `state_spec`)
        of the state class, it is directly returned.  Otherwise, we try to
        compute the quantity with the method :func:`compute`.

        It should not be necessary to redefine this method in child class.

        """

        if key in self.keys_state_spect:
            return self.state_spect.get_var(key)

        elif key in self.keys_state_phys:
            return self.state_phys.get_var(key)

        else:
            it = self.sim.time_stepping.it
            if key in self.vars_computed and it == self.it_computed[key]:
                return self.vars_computed[key]

            else:
                value = self.compute(key)
                self.vars_computed[key] = value
                self.it_computed[key] = it
                return value

    def __setitem__(self, key, value):
        """General setter function to set the value of a variable

        It should not be necessary to redefine this method in child class.
        """
        if key in self.keys_state_spect:
            self.state_spect.set_var(key, value)
        elif key in self.keys_state_phys:
            self.state_phys.set_var(key, value)
        else:
            raise ValueError('key "' + key + '" is not known')

    def statespect_from_statephys(self):
        """Compute the spectral variables from the physical variables.

        When you implement a new solver, check that this method does the job!
        """
        for ik in range(self.state_spect.nvar):
            self.oper.fft_as_arg(self.state_phys[ik], self.state_spect[ik])

    def statephys_from_statespect(self):
        """Compute the physical variables from the spectral variables.

        When you implement a new solver, check that this method does the job!
        """
        for ik in range(self.state_spect.nvar):
            self.oper.ifft_as_arg(self.state_spect[ik], self.state_phys[ik])

    def return_statephys_from_statespect(self, state_spect=None):
        """Return the physical variables computed from the spectral variables."""
        ifft = self.oper.ifft
        if state_spect is None:
            state_spect = self.state_spect

        state_phys = SetOfVariables(like=self.state_phys)
        for ik in range(self.state_spect.nvar):
            state_phys[ik] = ifft(state_spect[ik])
        return state_phys

    def can_this_key_be_obtained(self, key):
        return (
            key in self.keys_state_phys
            or key in self.keys_computable
            or key in self.keys_state_spect
        )

    def init_statespect_from(self, **kwargs):
        """Initialize `state_spect` from arrays.

        Parameters
        ----------

        **kwargs : {key: array, ...}

          keys and arrays used for the initialization. The other keys
          are set to zero.

        Examples
        --------

        .. code-block:: python

           kwargs = {'a_fft': Fa_fft}
           init_statespect_from(**kwargs)

           ux_fft, uy_fft, eta_fft = oper.uxuyetafft_from_qfft(q_fft)
           init_statespect_from(ux_fft=ux_fft, uy_fft=uy_fft, eta_fft=eta_fft)

        """
        self.state_spect[:] = 0.0

        for key, value in list(kwargs.items()):
            if key not in self.keys_state_spect:
                raise ValueError(
                    f'Do not know how to initialize with key "{key}".'
                )

            self.state_spect.set_var(key, value)



"""Utilities for the unittests (:mod:`fluidsim.util.testing`)
=============================================================

This module extends the functionalities provide by the `unittest` standard
library. It enforces matplotlib to use its `Agg` backend, thereby smoothly
executing plotting tests without window pop-ups. It is also useful to execute
this module to run FluidSim unittests, without going into the source directory
for eg. when it is installed through `pip`.

"""
import argparse
import inspect
import os
import shutil
import sys
import time
import unittest
from importlib import import_module
from warnings import warn

import numpy as np

from fluiddyn.io import stdout_redirected
from fluiddyn.util import mpi
from fluiddyn.util.compat import cached_property

try:
    import fluidfft
except ImportError:
    FLUIDFFT_INSTALLED = False
else:
    FLUIDFFT_INSTALLED = True


def skip_if_no_fluidfft(func):
    return unittest.skipUnless(FLUIDFFT_INSTALLED, "FluidFFT is not installed")(
        func
    )


class classproperty:
    """A combination of property and classmethod. Decorator that converts a
    method with a single cls argument into a property that can be accessed
    directly from the class.

    Borrowed from django.utils.functional

    License: BSD-3
    Copyright (c) Django Software Foundation and individual contributors.
    All rights reserved.

    """

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self


class TestCase(unittest.TestCase):

    # True except if pytest is used...
    has_to_redirect_stdout = not any(
        any(test_tool in arg for arg in sys.argv)
        for test_tool in ("pytest", "py.test")
    )

    def run(self, result=None):
        with stdout_redirected(self.has_to_redirect_stdout):
            super().run(result=result)


class TestSimul(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.init_params()
        with stdout_redirected(cls.has_to_redirect_stdout):
            cls.sim = cls.Simul(cls.params)

    @classmethod
    def tearDownClass(cls):
        # clean by removing the directory
        if mpi.rank == 0:
            if hasattr(cls, "sim"):
                shutil.rmtree(cls.sim.output.path_run, ignore_errors=True)

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        params.time_stepping.USE_CFL = False
        params.time_stepping.USE_T_END = False
        params.time_stepping.it_end = 2
        params.time_stepping.deltat0 = 0.1


class TestSimulConserve(TestSimul):
    """A test case which makes it easy to test for energy and enstrophy
    conservation. By default the simulation is instantiated and run till the end
    within the ``setUpClass`` method.

    """

    zero = 1e-14

    @classmethod
    def setUpClass(cls):
        cls.init_params()
        with stdout_redirected(cls.has_to_redirect_stdout), cls.Simul(
            cls.params
        ) as sim:
            cls.sim = sim
            sim.time_stepping.start()

    @cached_property
    def tendencies_fft(self):
        self.sim.params.forcing.enable = False
        return self.sim.tendencies_nonlin()

    def assertAlmostZero(
        self, value, places=None, msg=None, tolerance_warning=True
    ):
        """Assert if value is almost zero."""
        if places is None:
            places = -int(np.log10(self.zero))

        self.assertAlmostEqual(value, 0, places=places, msg=msg)
        if places < 7 and mpi.rank == 0 and tolerance_warning:
            warn(
                "Machine zero level too high. Value to be asserted as zero"
                f"= {value}\n\tTest: {self.id()}"
            )


class TestSimulConserveOutput(TestSimulConserve):
    """A test case with methods to easily test for the output modules."""

    def get_sim_output_attr_from_str(self, module):
        return getattr(self.sim.output, module)

    def get_results(self, name):
        module = self.get_sim_output_attr_from_str(name)
        for method_str in ("compute", "load_dataset", "load"):
            try:
                method = getattr(module, method_str)
            except AttributeError:
                pass
            else:
                results = method()
        return results

    @unittest.skipIf(mpi.nb_proc > 1, "plot function works sequentially only")
    def plot(self, name):
        """Test if plot methods work."""
        import matplotlib.pyplot as plt

        attr = self.get_sim_output_attr_from_str(name)
        attr.plot()
        plt.close("all")


class TimeLoggingTestResult(unittest.TextTestResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.test_timings = []

    def startTest(self, test):
        self._test_started_at = time.time()
        super().startTest(test)

    def addSuccess(self, test):
        elapsed = time.time() - self._test_started_at
        name = self.getDescription(test)
        self.test_timings.append((name, elapsed))
        super().addSuccess(test)

    def getTestTimings(self):
        return self.test_timings


class TimeLoggingTestRunner(unittest.TextTestRunner):
    def __init__(self, slow_test_threshold=0.3, *args, **kwargs):
        self.slow_test_threshold = slow_test_threshold
        return super().__init__(
            resultclass=TimeLoggingTestResult, *args, **kwargs
        )

    def run(self, tests):
        result = super().run(tests)
        msg = f"\n\nSlow tests (>{self.slow_test_threshold:.3f}s):"
        self.write_result(msg)
        self.write_result("-" * len(msg))

        for name, elapsed in result.getTestTimings():
            if elapsed > self.slow_test_threshold:
                self.write_result(f"({elapsed:.3f}s) {name}")

        return result

    def write_result(self, *strs):
        """Write strings to the result stream."""
        if mpi.rank == 0:
            msg = " ".join(strs)
            self.stream.writeln(msg)


def _mname(obj):
    """ Get the full dotted name of the test method """

    mod_name = obj.__class__.__module__.replace("fluidsim.", "")
    return f"{mod_name}.{obj.__class__.__name__}.{obj._testMethodName}"


def deactivate_redirect_stdout(tests):
    TestCase.has_to_redirect_stdout = False


def _run(tests, verbose=False, no_capture=False):
    """Run a set of tests using unittest."""

    if mpi.rank == 0:
        verbosity = 2
    else:
        verbosity = 0

    if verbose:
        TestRunner = TimeLoggingTestRunner
    else:
        TestRunner = unittest.runner.TextTestRunner

    testRunner = TestRunner(verbosity=verbosity)

    if no_capture:
        deactivate_redirect_stdout(tests)

    result = testRunner.run(tests)
    if verbose:
        msg = "Skipped tests"
        testRunner.write_result("\n", msg, "\n", "-" * len(msg))
        for (case, reason) in result.skipped:
            testRunner.write_result("S  {} ({})".format(_mname(case), reason))
        for (case, reason) in result.expectedFailures:
            testRunner.write_result("X  %s" % _mname(case))

    return result


def import_test_module(module_name: str):
    """Smarter import handling common mistakes with warnings."""
    if not module_name.startswith("fluidsim"):
        warn(
            f"Assuming you forgot to add fluidsim in front of module name: {module_name}"
        )
        module_name = ".".join(("fluidsim", module_name))

    try:
        module = import_module(module_name)
    except ModuleNotFoundError:
        module_name = ".".join(module_name.split(".")[:-1])
        # warn(f"Module not found. Attempting {module_name} instead")
        module = import_module(module_name)
    finally:
        return module


def discover_tests(
    verbose=False, no_capture=False, module_name="fluidsim", path=None
):
    """Discovers all tests under a module or directory.
    Similar to `python -m unittest discover`.

    Parameters
    ----------
    module_name: str

        Tests are discovered under this module

    start_dir: str

        Tests are discovered under this directory. Overrides `module_name`
        parameter.

    verbose: bool

        For verbose output

    """
    if path is None:
        module = import_test_module(module_name)
        path_src = inspect.getfile(module)
        path = os.path.dirname(path_src)

    if os.path.isdir(path):
        loader = unittest.TestLoader()
        tests = loader.discover(path)
    else:
        module = path.replace(os.path.sep, ".")
        module = import_test_module(module)
        tests = unittest.defaultTestLoader.loadTestsFromModule(module)
        suite = unittest.TestSuite()
        suite.addTests(tests)

    return _run(tests, verbose, no_capture)


def collect_tests(verbose, no_capture, *modules):
    """Creates a `TestSuite` from several modules.

    Parameters
    ----------
    modules: str, str, ...

        Strings representing modules containing atleast one unittest.TestCase
        class.

    Examples
    --------
    >>> collect_tests(
            'fluidsim.solvers.test.test_ns',
            'fluidsim.solvers.test.test_sw1l',
            'fluidsim.operators.test.test_operators2d')

    """
    suite = unittest.TestSuite()
    for module in modules:
        module = module.replace(os.path.sep, ".")
        module = import_test_module(module)
        tests = unittest.defaultTestLoader.loadTestsFromModule(module)
        suite.addTests(tests)

    return _run(suite, verbose, no_capture)


def init_parser(parser):
    """Initialize argument parser for `fluidsim-test`."""

    parser.add_argument(
        "-m",
        "--module",
        default="fluidsim",
        help="tests are discovered under this module",
    )
    parser.add_argument(
        "-p",
        "--path",
        default=None,
        help=(
            "tests are discovered under this directory/file and overrides -m"
            " option."
        ),
    )
    parser.add_argument(
        "-c",
        "--collect",
        nargs="+",
        default=[],
        help=(
            "create a TestSuite by collecting multiple modules containing"
            "TestCases"
        ),
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="for verbose output"
    )
    parser.add_argument(
        "-nc", "--no-capture", action="store_true", help="No stdout capture"
    )


def run(args=None):
    """Run fluidsim-test command."""

    if args is None:
        parser = argparse.ArgumentParser(
            prog="fluidsim-test",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="Run FluidSim tests using unittest from any directory.",
        )

        init_parser(parser)
        args = parser.parse_args()

    if len(args.collect) > 0:
        result = collect_tests(args.verbose, args.no_capture, *args.collect)
    else:
        result = discover_tests(
            args.verbose, args.no_capture, args.module, args.path
        )

    if result.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    run()



"""Solvers for a variety of physical problems
=============================================

Base package containing the source code of the particular solvers.  Some
sub-packages contains also solvers for variant of the mentioned equations (for
example, :mod:`fluidsim.solvers.ns2d.strat` or
:mod:`fluidsim.solvers.sw1l.modified`).

.. autosummary::
   :toctree:

   ns2d
   ns3d
   sw1l
   plate2d
   ad1d
   waves2d
   sphere
   models0d
   nl1d
   burgers1d

Provides:

.. autosummary::
   :toctree:

   pkgload

"""
import numpy as _np
from warnings import warn as _warn


def pkgload():
    """Populate ``fluidsim.solvers`` package namespace with the solvers for easier
    import.

    """
    from ..util.util import available_solver_keys

    solvers = available_solver_keys()

    solver_pkgs = (f"fluidsim.solvers.{solver}" for solver in solvers)
    solver_modules = (f"fluidsim.solvers.{solver}.solver" for solver in solvers)

    _warn(
        (
            "This function will not work in numpy>=0.16.0 and will be removed in "
            "the future"
        ),
        FutureWarning,
    )
    _np.pkgload(*solver_pkgs)
    _np.pkgload(*solver_modules)



import unittest

import matplotlib.pyplot as plt

import fluiddyn.util.mpi as mpi

import fluidsim as fls

from ..test_solver import TestSimulBase as _Base, classproperty


class TestSimulBase(_Base):
    @classproperty
    def Simul(self):
        from .solver import Simul

        return Simul


class TestOutput(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.init_fields.type = "dipole"

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.2

    def test_output(self):

        sim = self.sim
        sim.time_stepping.start()

        if mpi.nb_proc > 1:
            return

        sim2 = fls.load_sim_for_plot(sim.output.path_run)
        sim2.output.print_stdout.load()
        sim2.output.print_stdout.plot()
        sim2.output.spatial_means.load()
        sim2.output.spatial_means.load_dataset()
        sim2.output.spatial_means.plot()
        sim2.output.spectra.load1d_mean()
        sim2.output.spectra.load3d_mean()
        sim2.output.spectra.plot1d_times(
            tmin=0.1,
            tmax=10,
            delta_t=0.01,
            coef_compensate=5 / 3,
            coef_plot_k3=1.0,
            coef_plot_k53=1.0,
        )
        sim2.output.spectra.plot3d_times(
            tmin=0.1,
            tmax=10,
            delta_t=0.01,
            coef_compensate=5 / 3,
            coef_plot_k3=1.0,
            coef_plot_k53=1.0,
        )

        sim2.output.phys_fields.set_equation_crosssection(f"x={sim.oper.Lx/4}")
        sim2.output.phys_fields.animate("vx")

        sim2.output.phys_fields.plot(field="vx", time=10)

        plt.close("all")


class TestInitInScript(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.output.HAS_TO_SAVE = False
        params.init_fields.type = "in_script"

    def test_init_in_script(self):
        sim = self.sim
        # here we have to initialize the flow fields

        variables = {
            k: 1e-2 * sim.oper.create_arrayX_random() for k in ("vx", "vy", "vz")
        }

        X, Y, Z = sim.oper.get_XYZ_loc()

        sim.state.init_statephys_from(**variables)

        sim.state.statespect_from_statephys()
        sim.state.statephys_from_statespect()

        vx_fft = sim.state.get_var("vx_fft")
        vy_fft = sim.state.get_var("vy_fft")
        vz_fft = sim.state.get_var("vz_fft")

        sim.state.init_from_vxvyfft(vx_fft, vy_fft)
        sim.state.init_from_vxvyvzfft(vx_fft, vy_fft, vz_fft)


if __name__ == "__main__":
    unittest.main()



import unittest

import pytest

import numpy as np
import matplotlib.pyplot as plt

import fluiddyn.util.mpi as mpi

import fluidsim as fls

from fluidsim.extend_simul import extend_simul_class

from ..test_solver import TestSimulBase as _Base, classproperty


class TestSimulBase(_Base):
    @classproperty
    def Simul(cls):
        from .solver import Simul

        return Simul


class TestTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.oper.nx *= 2
        params.oper.ny *= 2
        params.oper.nz *= 2

        params.init_fields.type = "noise"
        params.output.HAS_TO_SAVE = False

    def test_tendency(self):

        sim = self.sim

        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)

        T_tot = np.ascontiguousarray(sim.oper.create_arrayK(value=0.0).real)

        for axis in ("x", "y", "z"):
            key = f"v{axis}_fft"
            vi_fft = sim.state.get_var(key)
            Fi_fft = tend.get_var(key)
            T_tot += np.real(Fi_fft.conj() * vi_fft)

        key = "b_fft"
        b_fft = sim.state.get_var(key)
        F_fft = tend.get_var(key)

        T_tot += 1.0 / sim.params.N ** 2 * np.real(F_fft.conj() * b_fft)

        ratio = sim.oper.sum_wavenumbers(T_tot) / sim.oper.sum_wavenumbers(
            abs(T_tot)
        )

        self.assertGreater(1e-15, abs(ratio))


class TestOutput(TestSimulBase):
    @classproperty
    def Simul(cls):
        from .solver import Simul
        from fluidsim.extend_simul.spatial_means_regions_milestone import (
            SpatialMeansRegions,
        )

        return extend_simul_class(Simul, SpatialMeansRegions)

    @classmethod
    def init_params(self):
        params = super().init_params()
        params.time_stepping.t_end = 0.3
        params.init_fields.type = "noise"

        params.forcing.enable = True
        params.forcing.type = "in_script"

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.2

        periods.spatiotemporal_spectra = 0.01

        periods.spatial_means = 0.05
        periods.spatial_means_regions = 0.05

        for tag in params.output._tag_children:
            if tag.startswith("periods"):
                continue
            child = params.output[tag]
            if hasattr(child, "HAS_TO_PLOT_SAVED"):
                child["HAS_TO_PLOT_SAVED"] = True

        params.output.spectra.kzkh_periodicity = 2

        params.output.spatiotemporal_spectra.probes_region = (
            params.oper.nx // 2,
            params.oper.ny // 2,
            params.oper.nz // 2,
        )

    @pytest.mark.filterwarnings("ignore:divide by zero encountered in log10")
    def test_output(self):

        sim = self.sim

        oper = sim.oper
        X, Y, Z = oper.get_XYZ_loc()

        def compute_forcing_fft_each_time(self):
            return {"vx_fft": oper.create_arrayK(value=0)}

        sim.forcing.forcing_maker.monkeypatch_compute_forcing_fft_each_time(
            compute_forcing_fft_each_time
        )

        sim.time_stepping.start()

        sim.state.compute("divh")

        np.testing.assert_almost_equal(
            sum(sim.output.compute_energies()), sim.output.compute_energy()
        )

        if mpi.nb_proc > 1:
            return

        sim2 = fls.load_sim_for_plot(sim.output.path_run)
        sim2.output.print_stdout.load()
        sim2.output.print_stdout.plot()
        sim2.output.print_stdout.plot_clock_times()

        data = sim2.output.spatial_means.load()
        assert "EKhs" in data
        sim2.output.spatial_means.load_dataset()
        sim2.output.spatial_means.plot(plot_injection=True, plot_hyper=True)
        sim2.output.spatial_means.plot_dt_E()

        sim2.output.spatial_means_regions.load()
        sim2.output.spatial_means_regions.plot()
        sim2.output.spatial_means_regions.plot_budget(
            iregion=1, decompose_fluxes=True, plot_conversion=True
        )

        sim2.output.spatial_means.plot_dimless_numbers_versus_time()
        result = sim2.output.spatial_means.get_dimless_numbers_averaged()
        assert isinstance(result, dict)

        sim2.output.spectra.load1d_mean()
        sim2.output.spectra.load3d_mean()
        sim2.output.spectra.load_kzkh_mean()
        sim2.output.spectra.plot1d()
        sim2.output.spectra.plot1d_times(
            tmin=0.1,
            tmax=10,
            delta_t=0.01,
            coef_compensate=5 / 3,
            coef_plot_k3=1.0,
            coef_plot_k53=1.0,
        )
        sim2.output.spectra.plot3d_times(
            tmin=0.1,
            tmax=10,
            delta_t=0.01,
            coef_compensate=5 / 3,
            coef_plot_k3=1.0,
            coef_plot_k53=1.0,
        )
        sim2.output.spectra.plot_kzkh(key="Khd")
        sim2.output.spectra.plot_kzkh_cumul_diss(tmin=0.1, tmax=10)

        sim2.output.spect_energy_budg.plot_kzkh()
        sim2.output.spect_energy_budg.plot_fluxes()

        sim2.output.phys_fields.set_equation_crosssection(f"x={sim.oper.Lx/4}")
        sim2.output.phys_fields.animate("vx")
        sim2.output.phys_fields.plot(field="vx", time=10)

        sim2.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="Kp", equation="omega=200 * N"
        )
        sim2.output.spatiotemporal_spectra.plot_temporal_spectra()

        plt.close("all")


class TestInitInScript(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.output.HAS_TO_SAVE = False
        params.init_fields.type = "in_script"

    def test_init_in_script(self):
        # here we have to initialize the flow fields
        sim = self.sim
        variables = {
            k: 1e-2 * sim.oper.create_arrayX_random() for k in ("vx", "vy", "vz")
        }

        X, Y, Z = sim.oper.get_XYZ_loc()

        sim.state.init_statephys_from(**variables)

        sim.state.statespect_from_statephys()
        sim.state.statephys_from_statespect()

        vx_fft = sim.state.get_var("vx_fft")
        vy_fft = sim.state.get_var("vy_fft")
        vz_fft = sim.state.get_var("vz_fft")

        sim.state.init_from_vxvyfft(vx_fft, vy_fft)
        sim.state.init_from_vxvyvzfft(vx_fft, vy_fft, vz_fft)


class TestNoShearModes(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.init_fields.type = "noise"

        params.forcing.enable = False

        params.output.periods_save.spatial_means = 0.2
        params.output.periods_save.spectra = 0.2
        params.output.spectra.kzkh_periodicity = 1

        params.oper.NO_SHEAR_MODES = True
        params.no_vz_kz0 = True

        # we need delta_kx == delta_ky for this test
        params.oper.Ly = params.oper.Lx
        params.oper.ny = params.oper.nx

    def test_noshearmodes(self):

        sim = self.sim

        sim.time_stepping.start()

        data = sim.output.spectra.load_kzkh_mean(
            tmin=0.2, key_to_load=["Khd", "Kz", "Khr", "A"]
        )
        means = sim.output.spatial_means.load()
        if mpi.nb_proc > 1:
            mpi.comm.barrier()

        # check k \cdot v = 0
        assert np.allclose(data["Khd"][0, :].sum(), 0.0)
        assert np.allclose(data["Kz"][:, 0].sum(), 0.0)

        # check energy in shear modes
        EKhs = means["EKhs"]
        E = means["E"]
        ratio = EKhs[-1] / E[-1]
        assert ratio < 1e-15

        # check energy in shear modes
        spectrum = data["Khr"] + data["Khd"] + data["Kz"] + data["A"]
        assert np.allclose(spectrum[:, 0].sum(), 0.0)

        # check energy in kz = 0
        spectrum = data["Khd"] + data["Kz"]
        assert np.allclose(spectrum[0, :].sum(), 0.0)


if __name__ == "__main__":
    unittest.main()



"""Information on a solver (:mod:`fluidsim.base.params`)
==============================================================


Provides:

.. autoclass:: Parameters
   :members:
   :private-members:


"""
from fluidsim_core.params import Parameters
from fluidsim.base.solvers.info_base import InfoSolverBase


def fix_old_params(params):
    """Fix old parameters with depreciated values."""
    # params.FORCING -> params.forcing.enable (2018-02-16)
    try:
        params.FORCING
    except AttributeError:
        pass
    else:
        try:
            params.forcing
        except AttributeError:
            pass
        else:
            params.forcing._set_attrib("enable", params.FORCING)

    # for ns2d.strat (wrong parameter params.NO_SHEAR_MODES)
    try:
        params.NO_SHEAR_MODES
    except AttributeError:
        pass
    else:
        try:
            params.oper.NO_SHEAR_MODES = params.NO_SHEAR_MODES
        except AttributeError:
            pass


def merge_params(to_params, *other_params):
    """Merges missing parameters attributes and children of a typical
    Simulation object's parameters when compared to other parameters.
    Also, tries to replace `to_params.oper.type_fft` if found to be
    not based on FluidFFT.

    Parameters
    ----------
    to_params: Parameters

    other_params: Parameters, Parameters, ...

    """
    for other in other_params:
        to_params |= other

    # Substitute old FFT types with newer FluidFFT implementations
    if hasattr(to_params, "oper") and hasattr(to_params.oper, "type_fft"):
        method = to_params.oper.type_fft
        if (
            not method
            or method != "default"
            and not any(
                [
                    method.startswith(prefix)
                    for prefix in ("fft2d.", "fft3d.", "fluidfft.")
                ]
            )
        ):
            type_fft = "default"
            print("params.oper.type_fft", to_params.oper.type_fft, "->", type_fft)
            to_params.oper.type_fft = type_fft


create_params = Parameters._create_params
load_params_simul = Parameters._load_params_simul
load_info_solver = Parameters._load_info_solver


if __name__ == "__main__":
    info_solver = InfoSolverBase(tag="solver")

    params = create_params(info_solver)

# info = create_info_simul(info_solver, params)



"""Initialisation of the fields (:mod:`fluidsim.solvers.sw1l.init_fields`)
==========================================================================

Provides:

.. autoclass:: InitFieldsNoise
   :members:
   :private-members:

.. autoclass:: InitFieldsWave
   :members:
   :private-members:

.. autoclass:: InitFieldsVortexGrid
   :members:
   :private-members:

.. autoclass:: InitFieldsSW1L
   :members:
   :private-members:

"""

import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.init_fields import InitFieldsBase, SpecificInitFields

from fluidsim.solvers.ns2d.init_fields import (
    InitFieldsNoise as InitFieldsNoiseNS2D,
)

from fluidsim.solvers.ns2d.init_fields import InitFieldsJet, InitFieldsDipole


class InitFieldsNoise(InitFieldsNoiseNS2D):
    def __call__(self):
        rot_fft, ux_fft, uy_fft = self.compute_rotuxuy_fft()
        self.sim.state.init_from_uxuyfft(ux_fft, uy_fft)


class InitFieldsWave(SpecificInitFields):
    tag = "wave"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)
        params.init_fields._set_child(cls.tag, attribs={"eta_max": 1.0, "ikx": 2})

    def __call__(self):
        oper = self.sim.oper

        ikx = self.sim.params.init_fields.wave.ikx
        eta_max = self.sim.params.init_fields.wave.eta_max

        kx = oper.deltakx * ikx
        eta_fft = np.zeros_like(self.sim.state.get_var("eta_fft"))
        cond = np.logical_and(oper.KX == kx, oper.KY == 0.0)
        eta_fft[cond] = eta_max
        oper.project_fft_on_realX(eta_fft)

        self.sim.state.init_from_etafft(eta_fft)


class InitFieldsVortexGrid(SpecificInitFields):
    """Initializes the vorticity field with n_vort^2 Gaussian vortices in a square
    grid.

    Vortices are randomly assigned clockwise / anti-clockwise directions;
    with equal no of vortices for each direction.

    Parameters
    ----------
    omega_max : Max vorticity of a single vortex at its center

    n_vort :

      No. of vortices along one edge of the square grid, should be even integer

    sd : Standard Deviation of the gaussian, optional
      If not specified, follows six-sigma rule based on half vortex spacing

    """

    tag = "vortex_grid"

    @classmethod
    def _complete_params_with_default(cls, params):
        super()._complete_params_with_default(params)
        params.init_fields._set_child(
            cls.tag, attribs={"omega_max": 1.0, "n_vort": 8, "sd": None}
        )

    def __call__(self):
        rot = self.vortex_grid_shape()
        rot_fft = self.sim.oper.fft2(rot)
        self.sim.state.init_from_rotfft(rot_fft)

    def vortex_grid_shape(self):
        oper = self.sim.oper
        params = self.sim.params.init_fields.vortex_grid

        Lx = oper.Lx
        Ly = oper.Ly
        XX = oper.XX
        YY = oper.YY
        N_vort = params.n_vort
        SD = params.sd

        if N_vort % 2 != 0:
            raise ValueError(
                "Cannot initialize a net circulation free field."
                "N_vort should be even."
            )

        dx_vort = Lx / N_vort
        dy_vort = Ly / N_vort
        x_vort = np.linspace(0, Lx, N_vort + 1) + dx_vort / 2.0
        y_vort = np.linspace(0, Ly, N_vort + 1) + dy_vort / 2.0
        sign_list = self._random_plus_minus_list()

        if SD is None:
            SD = min(dx_vort, dy_vort) / 12.0
            params.sd = SD

        amp = params.omega_max

        def wz_gaussian(x, y, sign):
            return sign * amp * np.exp(-(x ** 2 + y ** 2) / (2 * SD ** 2))

        omega = np.zeros(oper.shapeX_loc)
        for i in range(0, N_vort):
            x0 = x_vort[i]
            for j in range(0, N_vort):
                y0 = y_vort[j]
                sign = sign_list[i * N_vort + j]
                omega = omega + wz_gaussian(XX - x0, YY - y0, sign)

        return omega

    def _random_plus_minus_list(self):
        """
        Returns a list with of length n_vort^2, with equal number of pluses and minuses.
        """
        N = self.sim.params.init_fields.vortex_grid.n_vort ** 2
        if mpi.rank == 0:
            pm = np.ones(N)
            pm[::2] = -1
            np.random.shuffle(pm)
        else:
            pm = np.empty(N)

        if mpi.nb_proc > 1:
            from mpi4py.MPI import INT

            mpi.comm.Bcast([pm, INT])

        return pm


class InitFieldsSW1L(InitFieldsBase):
    """Init the fields for the solver SW1L."""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""

        InitFieldsBase._complete_info_solver(
            info_solver,
            classes=[
                InitFieldsNoise,
                InitFieldsJet,
                InitFieldsDipole,
                InitFieldsWave,
                InitFieldsVortexGrid,
            ],
        )



"""IPython / Jupyter magic commands
===================================

.. autoclass:: FluidsimMagics
   :members:
   :private-members:

"""
from fluiddyn.io.query import query_yes_no
from fluidsim_core.magic import MagicsCore
from IPython.core.magic import line_magic, magics_class
from IPython.core.magic_arguments import (
    argument,
    magic_arguments,
    parse_argstring,
)

from fluidsim.util.util import load_sim_for_plot, load_state_phys_file


@magics_class
class FluidsimMagics(MagicsCore):
    """Magic commands can be loaded in IPython or Jupyter as

    >>> %load_ext fluidsim.magic

    Examples
    --------

    - Magic command %fluidsim

    %fluidsim creates the variables `params` and `Simul` for a particular solver.

    Create default parameters for a solver:

    >>> %fluidsim ns2d

    If a variable `params` already exists, you will be ask if you really want to
    overwrite it. To skip this question:

    >>> %fluidsim ns2d -f

    List all available solvers and initialized simulation:

    >>> %fluidsim

    - Magic command %fluidsim_load

    %fluidsim_load creates the variables `sim`, `params` and `Simul` from an
    existing simulation.

    Load existing simulation excluding state_phys files:

    >>> %fluidsim_load

    Load existing simulation all options: force overwrite, with state_phys
    files, merging parameters:

    >>> %fluidsim_load -f -s -t -m

    - Other fluidsim magic commands

    Quick reference (print this help message):

    >>> %fluidsim_help

    Delete the objects `sim` and `params`:

    >>> %fluidsim_reset

    """

    entrypoint_grp = "fluidsim.solvers"

    @magic_arguments()
    @argument("-f", "--force-overwrite", action="store_true")
    @argument("-s", "--state-phys", action="store_true")
    @argument("-t", "--t-approx", type=float, default=None)
    @argument("-m", "--merge-missing-params", action="store_true")
    @argument(
        "directory",
        nargs="?",
        help="Optional: absolute path/relative path/name of directory",
        default=None,
    )
    @line_magic
    def fluidsim_load(self, line):
        args = parse_argstring(self.fluidsim_load, line)

        if not args.force_overwrite and (self.is_defined("sim")):
            if not query_yes_no(
                "The variables `sim` is defined in your user namespace. "
                "Do you really want to overwrite it?"
            ):
                return

        if args.state_phys:
            sim = load_state_phys_file(
                name_dir=args.directory,
                t_approx=args.t_approx,
                merge_missing_params=args.merge_missing_params,
            )
        else:
            sim = load_sim_for_plot(
                name_dir=args.directory,
                merge_missing_params=args.merge_missing_params,
            )

        user_ns = self.shell.user_ns
        user_ns["Simul"] = type(sim)
        user_ns["params"] = sim.params
        user_ns["sim"] = sim


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    fluidsim_magics = FluidsimMagics(ipython)
    ipython.register_magics(fluidsim_magics)



from warnings import warn
from ..output.phys_fields2d import PhysFieldsBase2D, MoviesBasePhysFields2D


class Movies2DSphericalHarmo(MoviesBasePhysFields2D):
    """Base class defining most functions for movies for 2D spherical data."""

    def _get_axis_data(self, shape=None):
        """Get axis data.

        Returns
        -------

        lons : array
          x-axis data.

        lats : array
          y-axis data.

        """
        if shape:
            warn("_get_axis_data: shape parameter ignored.")

        x = self.oper.get_grid1d_seq("lon")
        y = self.oper.get_grid1d_seq("lat")

        return x, y


class PhysFieldsSphericalHarmo(PhysFieldsBase2D):
    def _init_movies(self):
        self.movies = Movies2DSphericalHarmo(self.output, self)

    def _set_title(self, ax, key, time, vmax=None):
        title = (
            key
            + f", $t = {time:.3f}$, "
            + self.output.name_solver
            + fr", $l_\max = {self.params.oper.lmax:d}$"
        )
        if vmax is not None:
            title += r", $|\vec{v}|_{max} = $" + f"{vmax:.3f}"
        ax.set_title(title)

    def _compute_skip_quiver(self):
        # 4% of the Lx it is a great separation between vector arrows.
        lx = max(self.oper.lons)
        nx = len(self.oper.lons)
        delta_quiver = 0.04 * lx
        skip = (nx / lx) * delta_quiver
        skip = int(round(skip))
        if skip < 1:
            skip = 1
        return skip



"""Burgers1D solver (:mod:`fluidsim.solvers.burgers1d.solver`)
==============================================================

Provides:

.. autoclass:: Simul
   :members:
   :private-members:


"""
from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)
from fluidsim.base.state import StatePseudoSpectral


class StateBurgers1D(StatePseudoSpectral):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the pseudospectral solver Burgers 1d.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["u_fft"],
                "keys_state_phys": ["u"],
                "keys_computable": [],
                "keys_phys_needed": ["u"],
            }
        )


class InfoSolver(InfoSolverPseudoSpectral):
    def _init_root(self):
        super()._init_root()

        package = "fluidsim.solvers.burgers1d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "Burgers1D"

        classes = self.classes

        package_ad1d = "fluidsim.solvers.ad1d"

        classes.Operators.module_name = "fluidsim.operators.operators1d"
        classes.Operators.class_name = "OperatorsPseudoSpectral1D"

        classes.InitFields.module_name = package_ad1d + ".init_fields"
        classes.InitFields.class_name = "InitFieldsAD1D"

        classes.State.module_name = self.module_name
        classes.State.class_name = "StateBurgers1D"

        classes.Output.module_name = "fluidsim.solvers.ad1d.output"
        classes.Output.class_name = "Output"


class Simul(SimulBasePseudoSpectral):
    InfoSolver = InfoSolver

    def tendencies_nonlin(self, state_spect=None, old=None):

        if state_spect is None:
            u_fft = self.state.state_spect.get_var("u_fft")
            signal = self.state.state_phys.get_var("u")
        else:
            u_fft = state_spect.get_var("u_fft")
            signal = self.oper.ifft(u_fft)

        pxu_fft = self.oper.pxffft_from_fft(u_fft)
        pxu = self.oper.ifft(pxu_fft)

        f_signal = -signal * pxu

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        f_fft = tendencies_fft.get_var("u_fft")
        self.oper.fft_as_arg(f_signal, f_fft)
        self.oper.dealiasing(f_fft)
        # Set "oddball mode" to zero
        f_fft[self.oper.nkx-1] = 0.
        return tendencies_fft


if __name__ == "__main__":
    import numpy as np
    import matplotlib.pyplot as plt

    params = Simul.create_default_params()
    params.output.sub_directory = "examples"

    params.output.periods_save.phys_fields = 1.0
    params.short_name_type_run = "test"
    params.oper.Lx = 8
    params.oper.nx = 128
    params.oper.coef_dealiasing = 0.99
    params.nu_2 = 1e-2
    params.time_stepping.USE_CFL = False
    params.time_stepping.deltat0 = 1e-2
    params.time_stepping.t_end = 0.5
    params.time_stepping.type_time_scheme = "RK2"
    params.init_fields.type = "in_script"

    params.output.periods_print.print_stdout = 0.5

    sim = Simul(params)

    # Initialize
    x = sim.oper.x
    Lx = sim.oper.Lx

    u = np.zeros_like(x)

    coef = 4
    cond = x < Lx / 2
    u[cond] = np.tanh(coef * (x[cond] - Lx / 4))
    cond = x >= Lx / 2
    u[cond] = -np.tanh(coef * (x[cond] - 3 * Lx / 4))

    u_fft = sim.oper.fft(u)

    sim.state.init_statephys_from(u=u)
    sim.state.init_statespect_from(u_fft=u_fft)

    # Plot once
    sim.output.init_with_initialized_state()
    sim.output.phys_fields.plot(field="u")
    ax = plt.gca()

    sim.time_stepping.start()

    ax.plot(x, sim.state.state_phys.get_var("u"))
    plt.show()
